"""
Slime Invaders

Artwork from https://kenney.nl

This example shows how to:

* Get sprites to move as a group
* Change texture of sprites as a group
* Only have the bottom sprite in the group fire lasers
* Create 'shields' like in space invaders

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.slime_invaders
"""
import random
import arcade

SPRITE_SCALING_PLAYER = 0.75
SPRITE_SCALING_enemy = 0.75
SPRITE_SCALING_LASER = 1.0

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Slime Invaders"

BULLET_SPEED = 5
ENEMY_SPEED = 2

MAX_PLAYER_BULLETS = 3

# This margin controls how close the enemy gets to the left or right side
# before reversing direction.
ENEMY_VERTICAL_MARGIN = 15
RIGHT_ENEMY_BORDER = WINDOW_WIDTH - ENEMY_VERTICAL_MARGIN
LEFT_ENEMY_BORDER = ENEMY_VERTICAL_MARGIN

# How many pixels to move the enemy down when reversing
ENEMY_MOVE_DOWN_AMOUNT = 30

# Game state
GAME_OVER = 1
PLAY_GAME = 0


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = arcade.SpriteList()
        self.enemy_list = arcade.SpriteList()
        self.player_bullet_list = arcade.SpriteList()
        self.enemy_bullet_list = arcade.SpriteList()
        self.shield_list = arcade.SpriteList()

        # State of the game
        self.game_state = PLAY_GAME

        # Set up the player info
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING_PLAYER,
        )
        self.player_list.append(self.player_sprite)

        self.score = 0

        # Enemy movement
        self.enemy_change_x = -ENEMY_SPEED

        # Don't show the mouse cursor
        self.window.set_mouse_visible(False)

        # Load sounds. Sounds from kenney.nl
        self.gun_sound = arcade.load_sound(":resources:sounds/hurt5.wav")
        self.hit_sound = arcade.load_sound(":resources:sounds/hit5.wav")
        self.texture_enemy_left = arcade.load_texture(
            ":resources:images/enemies/slimeBlue.png",
        )
        self.texture_enemy_right = self.texture_enemy_left.flip_left_right()
        # The laser points right so we rotate it 270 clockwise to point up
        self.texture_blue_laser = arcade.load_texture(
            ":resources:images/space_shooter/laserBlue01.png",
        ).rotate_270()

        self.background_color = arcade.color.AMAZON
        self.score_text = arcade.Text("Score: 0", 10, 20, arcade.color.WHITE, 14)
        self.game_over_text = arcade.Text(
            "GAME OVER",
            x=self.width / 2,
            y=self.height / 2,
            color=arcade.color.WHITE,
            font_size=60,
            anchor_x="center",
        )

    def setup_level_one(self):
        # Load the textures for the enemies, one facing left, one right
        # Create rows and columns of enemies
        x_count = 7
        x_start = 380
        x_spacing = 80
        y_count = 5
        y_start = 470
        y_spacing = 60
        for x in range(x_start, x_spacing * x_count + x_start, x_spacing):
            for y in range(y_start, y_spacing * y_count + y_start, y_spacing):
                # Create the enemy instance
                # enemy image from kenney.nl
                enemy = arcade.Sprite(
                    self.texture_enemy_right,
                    scale=SPRITE_SCALING_enemy,
                    center_x=x,
                    center_y=y
                )
                # Add the enemy to the lists
                self.enemy_list.append(enemy)

    def make_shield(self, x_start):
        """
        Make a shield, which is just a 2D grid of solid color sprites
        stuck together with no margin so you can't tell them apart.
        """
        shield_block_width = 10
        shield_block_height = 20
        shield_width_count = 20
        shield_height_count = 5
        y_start = 150
        for x in range(x_start,
                       x_start + shield_width_count * shield_block_width,
                       shield_block_width):
            for y in range(y_start,
                           y_start + shield_height_count * shield_block_height,
                           shield_block_height):
                shield_sprite = arcade.SpriteSolidColor(shield_block_width,
                                                        shield_block_height,
                                                        color=arcade.color.WHITE)
                shield_sprite.center_x = x
                shield_sprite.center_y = y
                self.shield_list.append(shield_sprite)

    def reset(self):
        """
        Reset the game so it can be played again.
        This is not a standard Arcade method. It's simply an example of how
        you might reset the game.
        """
        self.game_state = PLAY_GAME

        # Clear the sprite lists
        self.enemy_list.clear()
        self.player_bullet_list.clear()
        self.enemy_bullet_list.clear()
        self.shield_list.clear()

        # Set up the player
        self.score = 0

        # Set default position for player
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 70

        # Make each of the shields
        step = self.width // 4 - 50
        print("make_shield", step)
        for x in [step, step * 2, step * 3]:
            print(x)
            self.make_shield(x)

        # Set the background color
        self.background_color = arcade.color.AMAZON

        self.setup_level_one()

    def on_draw(self):
        """Render the screen."""
        # Clear the window / screen with the configured background color
        self.clear()

        # Draw all the sprites.
        self.enemy_list.draw()
        self.player_bullet_list.draw()
        self.enemy_bullet_list.draw()
        self.shield_list.draw(pixelated=True)
        self.player_list.draw()

        # Update and draw the score
        self.score_text.text = f"Score: {self.score}"
        self.score_text.draw()

        # Draw game over if the game state is such
        if self.game_state == GAME_OVER:
            self.game_over_text.draw()
            self.window.set_mouse_visible(True)

    def on_key_press(self, key, modifiers):
        if key == arcade.key.ESCAPE:
            self.window.close()

    def on_mouse_motion(self, x, y, dx, dy):
        """
        Called whenever the mouse moves.
        """
        # Don't move the player if the game is over
        if self.game_state == GAME_OVER:
            return

        self.player_sprite.center_x = x

    def on_mouse_press(self, x, y, button, modifiers):
        """
        Called whenever the mouse button is clicked.
        """
        # Only allow the user so many bullets on screen at a time to prevent
        # them from spamming bullets.
        if len(self.player_bullet_list) < MAX_PLAYER_BULLETS:
            # Gunshot sound
            # arcade.play_sound(self.gun_sound)

            # Create a bullet
            bullet = arcade.Sprite(self.texture_blue_laser, scale=SPRITE_SCALING_LASER)

            # Give the bullet a speed
            bullet.change_y = BULLET_SPEED

            # Position the bullet
            bullet.center_x = self.player_sprite.center_x
            bullet.bottom = self.player_sprite.top

            # Add the bullet to the appropriate lists
            self.player_bullet_list.append(bullet)

    def update_enemies(self):
        # Move the enemy vertically
        for enemy in self.enemy_list:
            enemy.center_x += self.enemy_change_x

        # Check every enemy to see if any hit the edge. If so, reverse the
        # direction and flag to move down.
        move_down = False
        for enemy in self.enemy_list:
            if enemy.right > RIGHT_ENEMY_BORDER and self.enemy_change_x > 0:
                self.enemy_change_x *= -1
                move_down = True
            if enemy.left < LEFT_ENEMY_BORDER and self.enemy_change_x < 0:
                self.enemy_change_x *= -1
                move_down = True

        # Did we hit the edge above, and need to move t he enemy down?
        if move_down:
            # Yes
            for enemy in self.enemy_list:
                # Move enemy down
                enemy.center_y -= ENEMY_MOVE_DOWN_AMOUNT
                # Flip texture on enemy so it faces the other way
                if self.enemy_change_x > 0:
                    enemy.texture = self.texture_enemy_left
                else:
                    enemy.texture = self.texture_enemy_right

    def allow_enemies_to_fire(self):
        """
        See if any enemies will fire this frame.
        """
        # Track which x values have had a chance to fire a bullet.
        # Since enemy list is build from the bottom up, we can use
        # this to only allow the bottom row to fire.
        x_spawn = []
        for enemy in self.enemy_list:
            # Adjust the chance depending on the number of enemies. Fewer
            # enemies, more likely to fire.
            chance = 4 + len(self.enemy_list) * 4

            # Fire if we roll a zero, and no one else in this column has had
            # a chance to fire.
            if random.randrange(chance) == 0 and enemy.center_x not in x_spawn:
                # Create a bullet
                bullet = arcade.Sprite(
                    ":resources:images/space_shooter/laserRed01.png",
                    scale=SPRITE_SCALING_LASER,
                )

                # Angle down.
                bullet.angle = 180

                # Give the bullet a speed
                bullet.change_y = -BULLET_SPEED

                # Position the bullet so its top id right below the enemy
                bullet.center_x = enemy.center_x
                bullet.top = enemy.bottom

                # Add the bullet to the appropriate list
                self.enemy_bullet_list.append(bullet)

            # Ok, this column has had a chance to fire. Add to list so we don't
            # try it again this frame.
            x_spawn.append(enemy.center_x)

    def process_enemy_bullets(self):
        # Move the bullets
        self.enemy_bullet_list.update()

        # Loop through each bullet
        for bullet in self.enemy_bullet_list:
            # Check this bullet to see if it hit a shield
            hit_list = arcade.check_for_collision_with_list(bullet, self.shield_list)

            # If it did, get rid of the bullet and shield blocks
            if len(hit_list) > 0:
                bullet.remove_from_sprite_lists()
                for shield in hit_list:
                    shield.remove_from_sprite_lists()
                continue

            # See if the player got hit with a bullet
            if arcade.check_for_collision_with_list(self.player_sprite, self.enemy_bullet_list):
                self.game_state = GAME_OVER

            # If the bullet falls off the screen get rid of it
            if bullet.top < 0:
                bullet.remove_from_sprite_lists()

    def process_player_bullets(self):
        # Move the bullets
        self.player_bullet_list.update()

        # Loop through each bullet
        for bullet in self.player_bullet_list:

            # Check this bullet to see if it hit a enemy
            hit_list = arcade.check_for_collision_with_list(bullet, self.shield_list)
            # If it did, get rid of the bullet
            if len(hit_list) > 0:
                bullet.remove_from_sprite_lists()
                for shield in hit_list:
                    shield.remove_from_sprite_lists()
                continue

            # Check this bullet to see if it hit a enemy
            hit_list = arcade.check_for_collision_with_list(bullet, self.enemy_list)

            # If it did, get rid of the bullet
            if len(hit_list) > 0:
                bullet.remove_from_sprite_lists()

            # For every enemy we hit, add to the score and remove the enemy
            for enemy in hit_list:
                enemy.remove_from_sprite_lists()
                self.score += 1

                # Hit Sound
                arcade.play_sound(self.hit_sound)

            # If the bullet flies off-screen, remove it.
            if bullet.bottom > WINDOW_HEIGHT:
                bullet.remove_from_sprite_lists()

    def on_update(self, delta_time):
        """ Movement and game logic """
        if self.game_state == GAME_OVER:
            return

        self.update_enemies()
        self.allow_enemies_to_fire()
        self.process_enemy_bullets()
        self.process_player_bullets()

        if len(self.enemy_list) == 0:
            self.setup_level_one()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.reset()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from .texture import Texture, ImageData
from .spritesheet import SpriteSheet
from .loading import (
    load_texture,
    load_image,
    load_spritesheet,
)
from .generate import (
    make_circle_texture,
    make_soft_circle_texture,
    make_soft_square_texture,
)
from .tools import (
    get_default_texture,
    get_default_image,
)
from .manager import TextureCacheManager

default_texture_cache = TextureCacheManager()


__all__ = [
    "Texture",
    "ImageData",
    "load_texture",
    "load_image",
    "load_spritesheet",
    "make_circle_texture",
    "make_soft_circle_texture",
    "make_soft_square_texture",
    "get_default_texture",
    "get_default_image",
    "TextureCacheManager",
    "SpriteSheet",
    "default_texture_cache",
]



"""
An experimental splash screen for Arcade.

This is a simple splash screen that shows the Arcade logo
for a few seconds before the actual game starts.

If Arcade is properly installed, you can run this script with:
python -m arcade.gui.experimental.splash
"""

import arcade
from arcade import Sprite, SpriteList, Text, View


class ArcadeSplash(View):
    """This view shows an Arcade splash screen before the actual game starts.

    Args:
        view: Next view to show after the splash screen.
        duration: The duration of the splash screen in seconds. (Default 3 seconds)
        dark_mode: If True, the splash screen will be shown in dark mode. (Default False)
    """

    def __init__(self, view: View, duration: float = 1.5, dark_mode: bool = False):
        super().__init__()
        self._next_view = view
        self._duration = duration
        self._time = 0.0
        self._dark_mode = dark_mode

        _text_color = (255, 255, 255, 255) if dark_mode else (0, 0, 0, 255)
        self._bg_color = (0, 0, 0, 255) if dark_mode else arcade.color.WHITE_SMOKE

        self._sprites: SpriteList[Sprite] = SpriteList()
        self._logo = Sprite(
            arcade.load_texture(":system:/logo.png"),
            center_x=self.window.center_x,
            center_y=self.window.center_y,
        )
        self._logo.size = 300, 300
        self._sprites.append(self._logo)

        self._text = Text(
            "Python Arcade",
            anchor_x="center",
            x=self.window.center_x,
            y=self._logo.bottom - 20,
            color=_text_color,
            font_size=40,
            bold=True,
        )

    def on_show_view(self):
        """Set background color and reset time."""
        arcade.set_background_color(self._bg_color)
        self._time = 0.0
        self._logo.alpha = 0

    def on_update(self, delta_time: float):
        """Update the time and switch to the next view after the duration."""
        self._time += delta_time
        if self._time >= self._duration:
            self.window.show_view(self._next_view)

        # fade in Arcade logo
        self._logo.alpha = min(255, int(255 * self._time / self._duration))

    def on_draw(self):
        """Draw the sprites and text."""
        self.clear()
        self._sprites.draw()
        self._text.draw()


if __name__ == "__main__":
    window = arcade.Window()
    window.show_view(ArcadeSplash(ArcadeSplash(dark_mode=True, view=arcade.View())))
    arcade.run()



"""
Simple ray marcher.
This can also be created with shadertoy

To make this simpler to follow we've based it on the
"Ray Marching for Dummies!" video from The Art of Code
YouTube channel : https://www.youtube.com/watch?v=PGtv-dBi2wE
"""
from pathlib import Path
import arcade
from arcade.gl import geometry

CURRENT_DIR = Path(__file__).parent.resolve()


class RayMarcherSimple(arcade.Window):

    def __init__(self):
        super().__init__(720, 720, "Simple Ray Marcher", resizable=True)
        self.program = self.ctx.load_program(
            vertex_shader=CURRENT_DIR / "ray_marching_simple_vs.glsl",
            fragment_shader=CURRENT_DIR / "ray_marching_simple_fs.glsl"
        )
        self.quad_fs = geometry.quad_2d_fs()
        self.set_aspect_ratio(*self.get_size())

    def on_draw(self):
        self.quad_fs.render(self.program)

    def on_update(self, delta_time: float):
        self.program["iTime"] = self.time

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.set_aspect_ratio(width, height)

    def set_aspect_ratio(self, width, height):
        self.program["aspect_ratio"] = width / height


RayMarcherSimple().run()



"""

Example of a full functional inventory system.

This example demonstrates how to create a simple inventory system.

Main features are:
- Inventory slots
- Equipment slots
- Move items between slots
- Controller support

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.exp_controller_inventory
"""

# TODO: Drag and Drop

import pyglet.font
from pyglet.event import EVENT_HANDLED
from pyglet.gl import GL_NEAREST
from pyglet.input import Controller

import arcade
from arcade import Rect
from arcade.examples.gui.exp_controller_support_grid import (
    ControllerIndicator,
    setup_grid_focus_transition,
)
from arcade.experimental.controller_window import ControllerWindow, ControllerView
from arcade.gui import (
    Property,
    Surface,
    UIAnchorLayout,
    UIBoxLayout,
    UIFlatButton,
    UIGridLayout,
    UILabel,
    UIOnClickEvent,
    UIView,
    UIWidget,
    bind,
    UIEvent,
)
from arcade.gui.events import UIControllerButtonPressEvent
from arcade.gui.experimental.focus import UIFocusable, UIFocusGroup
from arcade.resources import load_kenney_fonts


class Item:
    """Base class for all items."""

    def __init__(self, symbol: str):
        self.symbol = symbol


class Inventory:
    """
    Basic inventory class.

    Contains items and manages items.


    inventory = Inventory(10)
    inventory.add(Item("🍎"))
    inventory.add(Item("🍌"))
    inventory.add(Item("🍇"))


    for item in inventory:
        print(item.symbol)

    inventory.remove(inventory[0])
    """

    def __init__(self, capacity: int):
        self._items: list[Item | None] = [None for _ in range(capacity)]
        self.capacity = capacity

    def add(self, item: Item):
        empty_slot = None
        for i, slot in enumerate(self._items):
            if slot is None:
                empty_slot = i
                break

        if empty_slot is not None:
            self._items[empty_slot] = item
        else:
            raise ValueError("Inventory is full.")

    def is_full(self):
        return len(self._items) == self.capacity

    def remove(self, item: Item):
        for i, slot in enumerate(self._items):
            if slot == item:
                self._items[i] = None
                return

    def __getitem__(self, index: int):
        return self._items[index]

    def __setitem__(self, index: int, value: Item):
        self._items[index] = value

    def __iter__(self):
        yield from self._items


class Equipment(Inventory):
    """Equipment inventory.

    Contains three slots for head, chest and legs.
    """

    def __init__(self):
        super().__init__(3)

    @property
    def head(self) -> Item:
        return self[0]

    @head.setter
    def head(self, value):
        self[0] = value

    @property
    def chest(self) -> Item:
        return self[1]

    @chest.setter
    def chest(self, value):
        self[1] = value

    @property
    def legs(self) -> Item:
        return self[2]

    @legs.setter
    def legs(self, value):
        self[2] = value


class InventorySlotUI(UIFocusable, UIFlatButton):
    """Represents a single inventory slot.
    The slot accesses a specific index in the inventory.

    Emits an on_click event.
    """

    def __init__(self, inventory: Inventory, index: int, **kwargs):
        super().__init__(size_hint=(1, 1), **kwargs)
        self.ui_label.update_font(font_size=24)
        self._inventory = inventory
        self._index = index

        item = inventory[index]
        if item:
            self.text = item.symbol

    @property
    def item(self) -> Item | None:
        return self._inventory[self._index]

    @item.setter
    def item(self, value):
        self._inventory[self._index] = value
        self._on_item_change()

    def _on_item_change(self, *args):
        if self.item:
            self.text = self.item.symbol
        else:
            self.text = ""


class EquipmentSlotUI(InventorySlotUI):
    pass


class InventoryUI(UIGridLayout):
    """Manages inventory slots.

    Emits an `on_slot_clicked(slot)` event when a slot is clicked.

    """

    def __init__(self, inventory: Inventory, **kwargs):
        super().__init__(
            size_hint=(0.7, 1),
            column_count=6,
            row_count=5,
            align_vertical="center",
            align_horizontal="center",
            vertical_spacing=10,
            horizontal_spacing=10,
            **kwargs,
        )
        self.with_padding(all=10)
        self.with_border(color=arcade.color.WHITE, width=2)

        self.inventory = inventory
        self.grid = {}

        for i, item in enumerate(inventory):
            slot = InventorySlotUI(inventory, i)
            # fill left to right, bottom to top (6x5 grid)
            self.add(slot, column=i % 6, row=i // 6)
            self.grid[(i % 6, i // 6)] = slot
            slot.on_click = self._on_slot_click  # type: ignore

        InventoryUI.register_event_type("on_slot_clicked")

    def _on_slot_click(self, event: UIOnClickEvent):
        # propagate slot click event to parent
        self.dispatch_event("on_slot_clicked", event.source)

    def on_slot_clicked(self, event: UIOnClickEvent):
        pass


class EquipmentUI(UIBoxLayout):
    """Contains three slots for equipment items.

    - Head
    - Chest
    - Legs

    Emits an `on_slot_clicked(slot)` event when a slot is clicked.

    """

    def __init__(self, **kwargs):
        super().__init__(size_hint=(0.3, 1), space_between=10, **kwargs)
        self.with_padding(all=20)
        self.with_border(color=arcade.color.WHITE, width=2)

        equipment = Equipment()

        self.head_slot = self.add(EquipmentSlotUI(equipment, 0))
        self.head_slot.on_click = lambda _: self.dispatch_event("on_slot_clicked", self.head_slot)

        self.chest_slot = self.add(EquipmentSlotUI(equipment, 1))
        self.chest_slot.on_click = lambda _: self.dispatch_event("on_slot_clicked", self.chest_slot)

        self.legs_slot = self.add(EquipmentSlotUI(equipment, 2))
        self.legs_slot.on_click = lambda _: self.dispatch_event("on_slot_clicked", self.legs_slot)

        EquipmentUI.register_event_type("on_slot_clicked")


class ActiveSlotTrackerMixin(UIWidget):
    """
    Mixin class to track the active slot.
    """

    active_slot = Property[InventorySlotUI | None](None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        bind(self, "active_slot", self.trigger_render)

    def do_render(self, surface: Surface):
        surface.limit(None)
        if self.active_slot:
            rect: Rect = self.active_slot.rect

            rect = rect.resize(*(rect.size + (2, 2)))
            arcade.draw_rect_outline(rect, arcade.uicolor.RED_ALIZARIN, 2)

        return super().do_render(surface)

    def on_slot_clicked(self, clicked_slot: InventorySlotUI):
        if self.active_slot:
            # disable active slot
            if clicked_slot == self.active_slot:
                self.active_slot = None
                return

            else:
                # swap items
                src_item = self.active_slot.item
                dst_item = clicked_slot.item

                self.active_slot.item = dst_item
                clicked_slot.item = src_item

                self.active_slot = None
                return

        else:
            # activate slot if contains item
            if clicked_slot.item:
                self.active_slot = clicked_slot


class InventoryModal(ActiveSlotTrackerMixin, UIFocusGroup, UIAnchorLayout):
    def __init__(self, inventory: Inventory, **kwargs):
        super().__init__(size_hint=(0.8, 0.8), **kwargs)
        self.with_padding(all=10)
        self.with_background(color=arcade.uicolor.GREEN_GREEN_SEA)
        self._debug = False

        self.add(
            UILabel(text="Inventory", font_size=20, font_name="Kenney Blocks", bold=True),
            anchor_y="top",
        )

        content = UIBoxLayout(size_hint=(1, 0.9), vertical=False, space_between=10)
        self.add(content, anchor_y="bottom")

        inv_ui = content.add(InventoryUI(inventory))
        inv_ui.on_slot_clicked = self.on_slot_clicked  # type: ignore

        eq_ui = content.add(EquipmentUI())
        eq_ui.on_slot_clicked = self.on_slot_clicked  # type: ignore

        # prepare focusable widgets
        widget_grid = inv_ui.grid
        setup_grid_focus_transition(
            widget_grid  # type: ignore
        )  # setup default transitions in a grid

        # add transitions to equipment slots
        cols = max(x for x, y in widget_grid.keys())
        rows = max(y for x, y in widget_grid.keys())

        equipment_slots = [eq_ui.head_slot, eq_ui.chest_slot, eq_ui.legs_slot]

        # connect inventory slots with equipment slots
        slots_to_eq_ratio = (rows + 1) / len(equipment_slots)
        for i in range(rows + 1):
            eq_index = int(i // slots_to_eq_ratio)
            eq_slot = equipment_slots[eq_index]

            inv_slot = widget_grid[(cols, i)]

            inv_slot.neighbor_right = eq_slot
            eq_slot.neighbor_left = inv_slot

        # close button not part of the normal focus rotation, but can be focused with "b"
        self.close_button = self.add(
            # todo: find out why X is not in center
            UIFlatButton(text="X", width=40, height=40),
            anchor_x="right",
            anchor_y="top",
        )
        self.close_button.on_click = lambda _: self.close()  # type: ignore

        # init controller support
        self.detect_focusable_widgets()

    def on_event(self, event: UIEvent) -> bool | None:
        if isinstance(event, UIControllerButtonPressEvent):
            if event.button == "b":
                self.set_focus(self.close_button)
                return EVENT_HANDLED

        return super().on_event(event)

    def close(self):
        self.visible = False
        self.trigger_full_render()


class MyView(UIView, ControllerView):
    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.BLACK

        self.inventory = Inventory(30)

        self.inventory.add(Item("🍎"))
        self.inventory.add(Item("🍌"))
        self.inventory.add(Item("🍇"))

        self.root = self.add_widget(UIAnchorLayout())
        self.add_widget(ControllerIndicator())

        text = self.root.add(
            UILabel(
                text="Open Inventory with 'Select' button on a controller or 'I' key", font_size=24
            )
        )
        text.fit_content()
        text.center_on_screen()

        self._inventory_modal = self.root.add(InventoryModal(self.inventory))

    def toggle_inventory(self):
        self._inventory_modal.visible = not self._inventory_modal.visible

    def on_key_press(self, symbol: int, modifiers: int) -> bool | None:
        if symbol == arcade.key.I:
            self.toggle_inventory()
            return True

        return super().on_key_press(symbol, modifiers)

    def on_button_press(self, controller: Controller, button):
        if button == "back":
            self.toggle_inventory()
            return True

        return super().on_button_press(controller, button)

    def on_draw_before_ui(self):
        pass


if __name__ == "__main__":
    # pixelate the font
    pyglet.font.base.Font.texture_min_filter = GL_NEAREST
    pyglet.font.base.Font.texture_mag_filter = GL_NEAREST

    load_kenney_fonts()

    ControllerWindow(title="Minimal example", width=1280, height=720, resizable=True).show_view(
        MyView()
    )
    arcade.run()



"""
Transform feedback example moving points using the mouse.
Each point as a desired position and a current position.
A radius around the mouse pointer will push the points
current position. Each point will constantly strive to
be at its desired position.

We're using a common trick ping-pong transforming
the point data between two buffers on the gpu
so we can keep working on the previous data.
This is mainly because it's not always safe to
read and write to the same buffer at the same time.

Increase the window resolution to get more points.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.transform_point_grid
"""

import random
from array import array

from pyglet.math import Mat4
import arcade
from arcade import gl
import arcade.clock


def gen_initial_data(window, width, height):
    """Generate a grid of points (current and desired position)"""
    dx, dy = window.width / width, window.height / height
    for y in range(height):
        for x in range(width):
            # current pos
            yield x * dx + dx / 2
            yield y * dy + dy / 2
            # desired pos
            yield x * dx + dx / 2
            yield y * dy + dy / 2


def gen_colors(width, height):
    """Generate some random colors"""
    for _ in range(width * height):
        yield random.uniform(0, 1)
        yield random.uniform(0, 1)
        yield random.uniform(0, 1)


class GameView(arcade.View):

    def __init__(self):
        super().__init__()
        self.window.set_vsync(True)
        self.ctx = self.window.ctx

        self.ctx.disable(self.ctx.BLEND)

        # Two buffers on the gpu with positions
        self.buffer1 = self.ctx.buffer(data=array("f", gen_initial_data(self, *self.size)))
        self.buffer2 = self.ctx.buffer(reserve=self.buffer1.size)
        # Buffer with color for each point
        self.colors = self.ctx.buffer(data=array("f", gen_colors(*self.size)))

        # Geometry for drawing the two buffer variants.
        # We pad away the desired position and add the color data.
        self.geometry1 = self.ctx.geometry(
            [
                gl.BufferDescription(self.buffer1, "2f 2x4", ["in_pos"]),
                gl.BufferDescription(self.colors, "3f", ["in_color"]),
            ]
        )
        self.geometry2 = self.ctx.geometry(
            [
                gl.BufferDescription(self.buffer2, "2f 2x4", ["in_pos"]),
                gl.BufferDescription(self.colors, "3f", ["in_color"]),
            ]
        )

        # Transform geometry for the two buffers. This is used to move the points
        # with a transform shader
        self.transform1 = self.ctx.geometry(
            [gl.BufferDescription(self.buffer1, "2f 2f", ["in_pos", "in_dest"])]
        )
        self.transform2 = self.ctx.geometry(
            [gl.BufferDescription(self.buffer2, "2f 2f", ["in_pos", "in_dest"])]
        )

        # Let's make the coordinate system match the viewport
        projection = Mat4.orthogonal_projection(0, self.width, 0, self.height, -100, 100)

        # Draw the points with the the supplied color
        self.points_program = self.ctx.program(
            vertex_shader="""
            #version 330

            uniform mat4 projection;
            in vec2 in_pos;
            in vec3 in_color;
            out vec3 color;

            void main() {
                gl_Position = projection * vec4(in_pos, 0.0, 1.0);
                color = in_color;
            }
            """,
            fragment_shader="""
            #version 330

            in vec3 color;
            out vec4 fragColor;

            void main() {
                fragColor = vec4(color, 1.0);
            }
            """,
        )
        # Write the projection matrix to the program uniform
        self.points_program["projection"] = projection

        # Program altering the point location.
        # We constantly try to move the point to its desired location.
        # In addition we check the distance to the mouse pointer and move it if
        # within a certain range.
        self.transform_program = self.ctx.program(
            vertex_shader="""
            #version 330

            uniform float dt; // time delta (between frames)
            uniform vec2 mouse_pos;

            in vec2 in_pos;
            in vec2 in_dest;

            out vec2 out_pos;
            out vec2 out_dest;

            void main() {
                out_dest = in_dest;
                // Slowly move the point towards the desired location
                vec2 dir = in_dest - in_pos;
                vec2 pos = in_pos + dir * dt;
                // Move the point away from the mouse position
                float dist = length(pos - mouse_pos);
                if (dist < 90.0) {
                    pos += (pos - mouse_pos) * dt * 10.0;
                }
                out_pos = pos;
            }
            """,
        )
        self.mouse_pos = -100, -100

    def on_draw(self):
        self.clear()
        # Set a point size. Multiply with pixel ratio taking retina resolutions
        # and UI scaling into account
        self.ctx.point_size = 2 * self.window.get_pixel_ratio()

        # Draw the points we calculated last frame
        self.geometry1.render(self.points_program, mode=gl.POINTS)

        # Move points with transform
        self.transform_program["dt"] = arcade.clock.GLOBAL_CLOCK.delta_time
        self.transform_program["mouse_pos"] = self.mouse_pos
        self.transform1.transform(self.transform_program, self.buffer2)

        # Swap variables around (ping-pong rendering)
        self.buffer1, self.buffer2 = self.buffer2, self.buffer1
        self.geometry1, self.geometry2 = self.geometry2, self.geometry1
        self.transform1, self.transform2 = self.transform2, self.transform1

    def on_mouse_motion(self, x, y, dx, dy):
        """Update the current mouse position"""
        self.mouse_pos = x, y


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(768, 768, "Moving Point Grid - GPU")

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Simple video player with cv2.
This player can be improved a lot, but it does a decent job.

We simply read frames from cv2 and write them into
a texture.

Dependencies:
    pip install opencv-python
"""

from math import floor
from pathlib import Path

import cv2  # type: ignore

import arcade
from arcade.gl.geometry import quad_2d_fs


class VideoPlayerCV2:
    """
    Primitive video player for Arcade with cv2.
    Renders to the entire screen. Use VideoPlayer to render to specific coordinate.

    Args:
        path: Path of the video that is to be played.
    """

    def __init__(self, path: str | Path, loop: bool = False):
        self.loop = loop

        self.ctx = arcade.get_window().ctx

        self.quad_fs = quad_2d_fs()
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_vert;
            in vec2 in_uv;
            out vec2 uv;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
                uv = in_uv;
            }
            """,
            fragment_shader="""
            #version 330

            uniform sampler2D videoFrame;
            out vec4 fragColor;
            in vec2 uv;

            void main() {
                fragColor = texture(videoFrame, vec2(uv.x, 1.0 - uv.y));
            }
            """,
        )

        # Configure videoFrame sampler to read from texture channel 0
        self.program["videoFrame"] = 0

        self.video = cv2.VideoCapture(str(arcade.resources.resolve(path)))

        # Query video size
        self._width, self._height = (
            int(self.video.get(cv2.CAP_PROP_FRAME_WIDTH)),
            int(self.video.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )

        # Get the framerate of the video
        self.video_frame_rate = self.video.get(cv2.CAP_PROP_FPS)
        # Keep track of the current frame and current time
        # to estimate a reasonable playback speed
        self.current_frame = 0
        self.time: float = 0.0

        # Get the number of frames in the video
        self.frames: int = self.video.get(cv2.CAP_PROP_FRAME_COUNT)  # type: ignore

        # Create and configure the OpenGL texture for the video
        self.texture = self.ctx.texture((self._width, self._height), components=3)
        # Swap the components in the texture because cv2 returns BGR data
        # Leave the alpha component as always 1
        self.texture.swizzle = "BGR1"

    @property
    def width(self) -> int:
        """Video width."""
        return self._width

    @property
    def height(self) -> int:
        """Video height."""
        return self._height

    def draw(self) -> None:
        """Call this in `on_draw`."""

        # Bind video texture to texture channel 0
        self.texture.use(unit=0)
        # Draw a fullscreen quad using our texture program
        self.quad_fs.render(self.program)

    def update(self, delta_time: float) -> None:
        """Move the frame forward."""
        self.time += delta_time

        # Do we need to read a new frame?
        next_frame = floor(self.time * self.video_frame_rate)
        if next_frame != self.current_frame:
            self.current_frame = next_frame
            # Read new frame from video and update texture
            exists, frame = self.video.read()
            if exists:
                self.texture.write(frame)
            # loop if we are at the end of the video
            elif self.loop:
                self.time = 0.0
                self.video.set(cv2.CAP_PROP_POS_FRAMES, 0)


class CV2PlayerView(arcade.View):
    """
    A simple view to hold a video player using cv2.

    Requires the opencv-python module to be installed.

    Args:
        path: Path of the video that is to be played.
        resize: Change the window size to the video size
    """

    def __init__(self, path: str | Path, loop: bool = False, resize: bool = False):
        super().__init__()

        self.video_player = VideoPlayerCV2(path, loop)

        if resize:
            self.window.set_size(self.video_player.width, self.video_player.height)

    def on_draw(self) -> None:
        self.clear()
        self.video_player.draw()

    def on_update(self, delta_time: float) -> None:
        self.video_player.update(delta_time)


if __name__ == "__main__":
    window = arcade.Window(1280, 720, "Video Player")
    window.show_view(CV2PlayerView(":resources:video/earth.mp4", loop=True, resize=False))
    window.run()



"""
Particle - Object produced by an Emitter.
Often used in large quantity to produce visual effects effects
"""

from typing import Literal

from arcade.math import clamp, lerp
from arcade.sprite import Sprite
from arcade.types import PathOrTexture, Point, Velocity


class Particle(Sprite):
    """Sprite that is emitted from an Emitter"""

    def __init__(
        self,
        path_or_texture: PathOrTexture | None,
        change_xy: tuple[float, float],
        center_xy: Point = (0.0, 0.0),
        angle: float = 0.0,
        change_angle: float = 0.0,
        scale: float = 1.0,
        alpha: int = 255,
        mutation_callback=None,
    ):
        super().__init__(path_or_texture, scale=scale)
        self.center_x = center_xy[0]
        self.center_y = center_xy[1]
        self.change_x = change_xy[0]
        self.change_y = change_xy[1]
        self.angle = angle
        self.change_angle = change_angle
        self.alpha = alpha
        self.mutation_callback = mutation_callback

    def update(self, delta_time: float = 1 / 60):
        """Advance the Particle's simulation"""
        super().update(delta_time)
        if self.mutation_callback:
            self.mutation_callback(self)

    # def draw(self):
    #     raise NotImplementedError("Particle.draw needs to be implemented")

    def can_reap(self):
        """Determine if Particle can be deleted"""
        raise NotImplementedError("Particle.can_reap needs to be implemented")


class EternalParticle(Particle):
    """Particle that has no end to its life"""

    def __init__(
        self,
        filename_or_texture: PathOrTexture,
        change_xy: Velocity,
        center_xy: Point = (0.0, 0.0),
        angle: float = 0,
        change_angle: float = 0,
        scale: float = 1.0,
        alpha: int = 255,
        mutation_callback=None,
    ):
        super().__init__(
            filename_or_texture,
            change_xy,
            center_xy,
            angle,
            change_angle,
            scale,
            alpha,
            mutation_callback,
        )

    def can_reap(self) -> Literal[False]:
        """Determine if Particle can be deleted"""
        return False


class LifetimeParticle(Particle):
    """Particle that lives for a given amount of time and is then deleted"""

    def __init__(
        self,
        filename_or_texture: PathOrTexture | None,
        change_xy: Velocity,
        lifetime: float,
        center_xy: Point = (0.0, 0.0),
        angle: float = 0,
        change_angle: float = 0,
        scale: float = 1.0,
        alpha: int = 255,
        mutation_callback=None,
    ):
        super().__init__(
            filename_or_texture,
            change_xy,
            center_xy,
            angle,
            change_angle,
            scale,
            alpha,
            mutation_callback,
        )
        self.lifetime_original = lifetime
        self.lifetime_elapsed = 0.0

    def update(self, delta_time: float = 1 / 60):
        """Advance the Particle's simulation"""
        super().update(delta_time)
        self.lifetime_elapsed += delta_time

    def can_reap(self) -> bool:
        """Determine if Particle can be deleted"""
        return self.lifetime_elapsed >= self.lifetime_original


class FadeParticle(LifetimeParticle):
    """Particle that animates its alpha between two values during its lifetime"""

    def __init__(
        self,
        filename_or_texture: PathOrTexture,
        change_xy: Velocity,
        lifetime: float,
        center_xy: Point = (0.0, 0.0),
        angle: float = 0,
        change_angle: float = 0,
        scale: float = 1.0,
        start_alpha: int = 255,
        end_alpha: int = 0,
        mutation_callback=None,
    ):
        super().__init__(
            filename_or_texture,
            change_xy,
            lifetime,
            center_xy,
            angle,
            change_angle,
            scale,
            start_alpha,
            mutation_callback,
        )
        self.start_alpha = start_alpha
        self.alpha = start_alpha
        self.end_alpha = end_alpha

    def update(self, delta_time: float = 1 / 60):
        """Advance the Particle's simulation"""
        super().update(delta_time)
        a = lerp(self.start_alpha, self.end_alpha, self.lifetime_elapsed / self.lifetime_original)
        self.alpha = int(clamp(a, 0, 255))



"""
City Scape Generator

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.shape_list_demo_skylines
"""
import random
import arcade
from arcade.shape_list import (
    ShapeElementList,
    create_rectangle_filled,
    create_polygon,
    create_rectangles_filled_with_colors,
)

WINDOW_WIDTH = 1200
WINDOW_HEIGHT = 600
WINDOW_TITLE = "Skyline Using Buffered Shapes"


def make_star_field(star_count):
    """Make a bunch of circles for stars"""
    shape_list = ShapeElementList()

    for _ in range(star_count):
        x = random.randrange(WINDOW_WIDTH)
        y = random.randrange(WINDOW_HEIGHT)
        radius = random.randrange(1, 4)
        brightness = random.randrange(127, 256)
        color = (brightness, brightness, brightness)
        shape = create_rectangle_filled(x, y, radius, radius, color)
        shape_list.append(shape)

    return shape_list


def make_skyline(width, skyline_height, skyline_color,
                 gap_chance=0.70, window_chance=0.30, light_on_chance=0.5,
                 window_color=(255, 255, 200), window_margin=3, window_gap=2,
                 cap_chance=0.20):
    """Make a skyline of buildings"""
    shape_list = ShapeElementList()

    # Add the "base" that we build the buildings on
    shape = create_rectangle_filled(
        center_x=width / 2,
        center_y=skyline_height / 2,
        width=width,
        height=skyline_height,
        color=skyline_color,
    )
    shape_list.append(shape)

    building_center_x = 0

    skyline_point_list = []
    color_list = []

    while building_center_x < width:

        # Is there a gap between the buildings?
        if random.random() < gap_chance:
            gap_width = random.randrange(10, 50)
        else:
            gap_width = 0

        # Figure out location and size of building
        building_width = random.randrange(20, 70)
        building_height = random.randrange(40, 150)
        building_center_x += gap_width + (building_width / 2)
        building_center_y = skyline_height + (building_height / 2)

        x1 = building_center_x - building_width / 2
        x2 = building_center_x + building_width / 2
        y1 = skyline_height
        y2 = skyline_height + building_height

        skyline_point_list.append([x1, y1])
        skyline_point_list.append([x1, y2])
        skyline_point_list.append([x2, y2])
        skyline_point_list.append([x2, y1])

        for i in range(4):
            color_list.append([skyline_color[0], skyline_color[1], skyline_color[2]])

        if random.random() < cap_chance:
            x1 = building_center_x - building_width / 2
            x2 = building_center_x + building_width / 2
            x3 = building_center_x

            y1 = y2 = building_center_y + building_height / 2
            y3 = y1 + building_width / 2

            # Roof
            shape = create_polygon([[x1, y1], [x2, y2], [x3, y3]], skyline_color)
            shape_list.append(shape)

        # See if we should have some windows
        if random.random() < window_chance:
            # Yes windows! How many windows?
            window_rows = random.randrange(10, 15)
            window_columns = random.randrange(1, 7)

            # Based on that, how big should they be?
            window_height = (building_height - window_margin * 2) / window_rows
            window_width = (
                (building_width - window_margin * 2 - window_gap * (window_columns - 1))
                / window_columns
            )

            # Find the bottom left of the building so we can start adding widows
            building_base_y = building_center_y - building_height / 2
            building_left_x = building_center_x - building_width / 2

            # Loop through each window
            for row in range(window_rows):
                for column in range(window_columns):
                    if random.random() > light_on_chance:
                        continue

                    x1 = (
                        building_left_x
                        + column * (window_width + window_gap)
                        + window_margin
                    )
                    x2 = (
                        building_left_x
                        + column * (window_width + window_gap)
                        + window_width
                        + window_margin
                    )
                    y1 = building_base_y + row * window_height
                    y2 = building_base_y + row * window_height + window_height * .8

                    skyline_point_list.append([x1, y1])
                    skyline_point_list.append([x1, y2])
                    skyline_point_list.append([x2, y2])
                    skyline_point_list.append([x2, y1])

                    for i in range(4):
                        color_list.append((
                            window_color[0],
                            window_color[1],
                            window_color[2],
                        ))

        building_center_x += (building_width / 2)

    shape = create_rectangles_filled_with_colors(skyline_point_list, color_list)
    shape_list.append(shape)

    return shape_list


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()
        # Enable vertical sync to make scrolling smoother
        self.window.set_vsync(True)

        self.stars = make_star_field(150)
        self.skyline1 = make_skyline(WINDOW_WIDTH * 5, 250, (80, 80, 80))
        self.skyline2 = make_skyline(WINDOW_WIDTH * 5, 150, (50, 50, 50))

        self.background_color = arcade.color.BLACK

    def on_draw(self):
        """Draw to screen"""
        self.clear()

        self.stars.draw()
        self.skyline1.draw()
        self.skyline2.draw()

    def on_update(self, delta_time):
        """Per frame update logic"""
        # Scroll each shape list with a slight offset to give a parallax effect
        self.skyline1.center_x -= 0.5 * 60 * delta_time
        self.skyline2.center_x -= 1 * 60 * delta_time

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int):
        """Make it possible scroll the scene around by dragging the mouse"""
        self.skyline1.center_x += dx
        self.skyline1.center_y += dy

        self.skyline2.center_x += dx
        self.skyline2.center_y += dy


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from arcade.gui.constructs import UIMessageBox, UIButtonRow
from arcade.gui.events import UIEvent
from arcade.gui.events import UIKeyEvent
from arcade.gui.events import UIKeyPressEvent
from arcade.gui.events import UIKeyReleaseEvent
from arcade.gui.events import UIMouseDragEvent
from arcade.gui.events import UIMouseEvent
from arcade.gui.events import UIMouseMovementEvent
from arcade.gui.events import UIMousePressEvent
from arcade.gui.events import UIMouseReleaseEvent
from arcade.gui.events import UIMouseScrollEvent
from arcade.gui.events import UIOnActionEvent
from arcade.gui.events import UIOnChangeEvent
from arcade.gui.events import UIOnClickEvent
from arcade.gui.events import UIOnUpdateEvent
from arcade.gui.events import UITextEvent
from arcade.gui.events import UITextInputEvent
from arcade.gui.events import UITextMotionEvent
from arcade.gui.events import UITextMotionSelectEvent
from arcade.gui.mixins import UIDraggableMixin
from arcade.gui.mixins import UIMouseFilterMixin
from arcade.gui.mixins import UIWindowLikeMixin
from arcade.gui.nine_patch import NinePatchTexture
from arcade.gui.property import ListProperty, DictProperty, Property, bind, unbind
from arcade.gui.style import UIStyleBase, UIStyledWidget
from arcade.gui.surface import Surface
from arcade.gui.ui_manager import UIManager
from arcade.gui.widgets import UIDummy
from arcade.gui.widgets import UIInteractiveWidget
from arcade.gui.widgets import UILayout
from arcade.gui.widgets import UISpace
from arcade.gui.view import UIView
from arcade.gui.widgets.dropdown import UIDropdown
from arcade.gui.widgets import UISpriteWidget
from arcade.gui.widgets import UIWidget
from arcade.gui.widgets.buttons import (
    UITextureButton,
    UITextureButtonStyle,
    UIFlatButton,
)
from arcade.gui.widgets.image import UIImage
from arcade.gui.widgets.layout import UIBoxLayout, UIAnchorLayout, UIGridLayout
from arcade.gui.widgets.slider import UIBaseSlider, UISlider, UITextureSlider, UISliderStyle
from arcade.gui.widgets.text import UILabel, UIInputText, UITextArea, UITextWidget
from arcade.gui.widgets.toggle import UITextureToggle

__all__ = [
    "UIAnchorLayout",
    "UIBoxLayout",
    "UIButtonRow",
    "UIGridLayout",
    "UIManager",
    "UIMessageBox",
    "UIKeyEvent",
    "UIDummy",
    "UIDraggableMixin",
    "UIDropdown",
    "UIMouseFilterMixin",
    "UIWindowLikeMixin",
    "UIKeyPressEvent",
    "UIKeyReleaseEvent",
    "UIEvent",
    "UIFlatButton",
    "UIImage",
    "UIInteractiveWidget",
    "UIInputText",
    "UILayout",
    "UILabel",
    "UIView",
    "UIMouseEvent",
    "UIMouseDragEvent",
    "UIMouseMovementEvent",
    "UIMousePressEvent",
    "UIMouseReleaseEvent",
    "UIMouseScrollEvent",
    "UIOnUpdateEvent",
    "UIOnActionEvent",
    "UIOnChangeEvent",
    "UIOnClickEvent",
    "UIBaseSlider",
    "UISlider",
    "UITextureSlider",
    "UISliderStyle",
    "UIStyleBase",
    "UIStyledWidget",
    "UISpace",
    "UISpriteWidget",
    "UITextArea",
    "UITextEvent",
    "UITextInputEvent",
    "UITextMotionEvent",
    "UITextMotionSelectEvent",
    "UITextureButton",
    "UITextureButtonStyle",
    "UITextureToggle",
    "UITextWidget",
    "UIWidget",
    "Surface",
    "NinePatchTexture",
    # Property classes
    "ListProperty",
    "DictProperty",
    "Property",
    "bind",
    "unbind",
]



from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, TypeVar

import arcade
from arcade.color import BLACK, WHITE
from arcade.exceptions import ReplacementWarning, warning
from arcade.hitbox import HitBox
from arcade.texture import Texture
from arcade.types import LRBT, AsFloat, Color, Point, Point2, Point2List, Rect, RGBOrA255
from arcade.utils import copy_dunders_unimplemented

if TYPE_CHECKING:
    from arcade.sprite_list import SpriteList

# Type from sprite that can be any BasicSprite or any subclass of BasicSprite
SpriteType = TypeVar("SpriteType", bound="BasicSprite")

# Same as SpriteType, for covariant type parameters
SpriteType_co = TypeVar("SpriteType_co", bound="BasicSprite", covariant=True)


@copy_dunders_unimplemented  # See https://github.com/pythonarcade/arcade/issues/2074
class BasicSprite:
    """
    The absolute minimum needed for a sprite.

    It does not support features like rotation or changing the hitbox
    after creation. For more built-in features, please see
    :py:class:`~arcade.Sprite`.

    Args:
        texture:
            The texture data to use for this sprite.
        scale:
            The scaling factor for drawing the texture.
        center_x:
            Location of the sprite along the X axis in pixels.
        center_y:
            Location of the sprite along the Y axis in pixels.
    """

    __slots__ = (
        "_position",
        "_depth",
        "_width",
        "_height",
        "_scale",
        "_color",
        "_texture",
        "_hit_box",
        "_visible",
        "sprite_lists",
        "_angle",
        "__weakref__",
    )

    def __init__(
        self,
        texture: Texture,
        scale: float | Point2 = 1.0,
        center_x: float = 0,
        center_y: float = 0,
        visible: bool = True,
        **kwargs: Any,
    ) -> None:
        self._position = (center_x, center_y)
        self._depth = 0.0
        self._texture = texture
        width, height = texture.size
        self._scale = (scale, scale) if isinstance(scale, float | int) else (scale[0], scale[1])  # noqa: UP038
        self._width = width * self._scale[0]
        self._height = height * self._scale[1]
        self._visible = bool(visible)
        self._color: Color = WHITE

        # In a more powerful type system, this would be typed as
        # list[SpriteList[? super Self]]
        # i.e., a list of SpriteList's with varying type arguments, but where
        # each of those type arguments is known to be a supertype of Self.
        # All changes to this list should go through the pair of methods
        # register_sprite_list, _unregister_sprite_list.
        # They ensure that the above typing invariant is preserved.
        self.sprite_lists: list[SpriteList[Any]] = []
        """The sprite lists this sprite is a member of"""

        # Core properties we don't use, but spritelist expects it
        self._angle = 0.0

        self._hit_box = HitBox(self._texture.hit_box_points, self._position, self._scale)

    # --- Core Properties ---

    @property
    def position(self) -> Point2:
        """Get or set the center x and y position of the sprite."""
        return self._position

    @position.setter
    def position(self, new_value: Point2):
        if new_value == self._position:
            return

        self._position = new_value
        self._hit_box.position = new_value
        self.update_spatial_hash()

        for sprite_list in self.sprite_lists:
            sprite_list._update_position(self)

    @property
    def center_x(self) -> float:
        """Get or set the center x position of the sprite."""
        return self._position[0]

    @center_x.setter
    def center_x(self, new_value: float):
        if new_value == self._position[0]:
            return

        self.position = (new_value, self._position[1])

    @property
    def center_y(self) -> float:
        """Get or set the center y position of the sprite."""
        return self._position[1]

    @center_y.setter
    def center_y(self, new_value: float):
        if new_value == self._position[1]:
            return

        self.position = (self._position[0], new_value)

    @property
    def depth(self) -> float:
        """
        Get or set the depth of the sprite.

        This is really the z coordinate of the sprite
        and can be used with OpenGL depth testing with opaque
        sprites.
        """
        return self._depth

    @depth.setter
    def depth(self, new_value: float):
        if new_value != self._depth:
            self._depth = new_value
            for sprite_list in self.sprite_lists:
                sprite_list._update_depth(self)

    @property
    def width(self) -> float:
        """Get or set width or the sprite in pixels"""
        return self._width

    @width.setter
    def width(self, new_value: float):
        if new_value != self._width:
            self._scale = new_value / self._texture.width, self._scale[1]
            self._hit_box.scale = self._scale
            self._width = new_value

            self.update_spatial_hash()
            for sprite_list in self.sprite_lists:
                sprite_list._update_width(self)

    @property
    def height(self) -> float:
        """Get or set the height of the sprite in pixels."""
        return self._height

    @height.setter
    def height(self, new_value: float):
        if new_value != self._height:
            self._scale = self._scale[0], new_value / self._texture.height
            self._hit_box.scale = self._scale
            self._height = new_value

            self.update_spatial_hash()
            for sprite_list in self.sprite_lists:
                sprite_list._update_height(self)

    @property
    def size(self) -> Point:
        """
        Get or set the size of the sprite as a pair of values.

        This is faster than getting or setting width and height separately.
        """
        return self._width, self._height

    @size.setter
    def size(self, new_value: Point2):
        try:
            width, height = new_value
        except ValueError:
            raise ValueError(
                "size must be a tuple-like object which unpacks to exactly 2 coordinates"
            )
        except TypeError:
            raise TypeError(
                "size must be a tuple-like object which unpacks to exactly 2 coordinates"
            )

        if width != self._width or height != self._height:
            texture_width, texture_height = self._texture.size
            self._scale = width / texture_width, height / texture_height
            self._hit_box.scale = self._scale
            self._width = width
            self._height = height

            self.update_spatial_hash()

            for sprite_list in self.sprite_lists:
                sprite_list._update_size(self)

    @property
    def scale_x(self) -> float:
        """
        Get or set the sprite's x scale value.

        .. note:: Negative values are supported. They will flip &
                  mirror the sprite.
        """
        return self._scale[0]

    @scale_x.setter
    def scale_x(self, new_scale_x: AsFloat):
        old_scale_x, old_scale_y = self._scale
        if new_scale_x == old_scale_x:
            return

        new_scale = (new_scale_x, old_scale_y)

        # Apply scale to hitbox first to raise any exceptions quickly
        self._hit_box.scale = new_scale
        self._scale = new_scale
        self._width = self._texture.width * new_scale_x

        self.update_spatial_hash()
        for sprite_list in self.sprite_lists:
            sprite_list._update_size(self)

    @property
    def scale_y(self) -> float:
        """
        Get or set the sprite's y scale value.

        .. note:: Negative values are supported. They will flip &
                  mirror the sprite.
        """
        return self._scale[1]

    @scale_y.setter
    def scale_y(self, new_scale_y: AsFloat):
        old_scale_x, old_scale_y = self._scale
        if new_scale_y == old_scale_y:
            return

        new_scale = (old_scale_x, new_scale_y)

        # Apply scale to hitbox first to raise any exceptions quickly
        self._hit_box.scale = new_scale
        self._scale = new_scale
        self._height = self._texture.height * new_scale_y

        self.update_spatial_hash()
        for sprite_list in self.sprite_lists:
            sprite_list._update_size(self)

    @property
    def scale(self) -> Point2:
        """Get or set the x & y scale of the sprite as a pair of values.
        You may set both the x & y with a single scalar, but scale will always return
        a length 2 tuple of the x & y scale

        See :py:attr:`.scale_x` and :py:attr:`.scale_y` for individual access.

        See :py:meth:`.scale_multiply_uniform` for uniform scaling.

        .. note:: Negative scale values are supported.

            This applies to both single-axis and dual-axis.
            Negatives will flip & mirror the sprite, but the
            with will use :py:func:`abs` to report total width
            and height instead of negatives.

        """
        return self._scale

    @scale.setter
    def scale(self, new_scale: Point2 | AsFloat):
        if isinstance(new_scale, float | int):
            scale_x = new_scale
            scale_y = new_scale

        else:  # Treat it as some sort of iterable or sequence
            # Don't abstract this. Keep it here since it's a hot code path
            try:
                scale_x, scale_y = new_scale  # type / length implicit check
            except ValueError:
                raise ValueError(
                    "scale must be a tuple-like object which unpacks to exactly 2 coordinates"
                )
            except TypeError:
                raise TypeError(
                    "scale must be a tuple-like object which unpacks to exactly 2 coordinates"
                )

        new_scale = scale_x, scale_y
        if new_scale == self._scale:
            return

        self._hit_box.scale = new_scale
        tex_width, tex_height = self._texture.size
        self._scale = new_scale
        self._width = tex_width * scale_x
        self._height = tex_height * scale_y

        self.update_spatial_hash()
        for sprite_list in self.sprite_lists:
            sprite_list._update_size(self)

    @property
    def left(self) -> float:
        """
        The leftmost x coordinate in the hit box.

        When setting this property the sprite is positioned
        relative to the leftmost x coordinate in the hit box.
        """
        return self._hit_box.left

    @left.setter
    def left(self, amount: float):
        leftmost = self.left
        diff = amount - leftmost
        self.center_x += diff

    @property
    def right(self) -> float:
        """
        The rightmost x coordinate in the hit box.

        When setting this property the sprite is positioned
        relative to the rightmost x coordinate in the hit box.
        """
        return self._hit_box.right

    @right.setter
    def right(self, amount: float):
        rightmost = self.right
        diff = rightmost - amount
        self.center_x -= diff

    @property
    def bottom(self) -> float:
        """
        The lowest y coordinate in the hit box.

        When setting this property the sprite is positioned
        relative to the lowest y coordinate in the hit box.
        """
        return self._hit_box.bottom

    @bottom.setter
    def bottom(self, amount: float):
        lowest = self.bottom
        diff = lowest - amount
        self.center_y -= diff

    @property
    def top(self) -> float:
        """
        The highest y coordinate in the hit box.

        When setting this property the sprite is positioned
        relative to the highest y coordinate in the hit box.
        """
        return self._hit_box.top

    @top.setter
    def top(self, amount: float):
        highest = self.top
        diff = highest - amount
        self.center_y -= diff

    @property
    def rect(self) -> Rect:
        """A rectangle with with the sprites left, right, bottom, and top values."""
        return LRBT(self.left, self.right, self.bottom, self.top)

    @property
    def visible(self) -> bool:
        """Get or set the visibility of this sprite.

        When set to ``False``, each :py:class:`~arcade.SpriteList` and
        its attached shaders will treat the sprite as if has an
        :py:attr:`.alpha` of 0. However, the sprite's actual values for
        :py:attr:`.alpha` and :py:attr:`.color` will not change.

        .. code-block:: python

            # The initial color of the sprite
            >>> sprite.color
            Color(255, 255, 255, 255)

            # Make the sprite invisible
            >>> sprite.visible = False
            # The sprite's color value has not changed
            >>> sprite.color
            Color(255, 255, 255, 255)
            # The sprite's alpha value hasn't either
            >>> sprite.alpha
            255

            # Restore visibility
            >>> sprite.visible = True
            # Shorthand to toggle visible
            >>> sprite.visible = not sprite.visible

        """
        return self._visible

    @visible.setter
    def visible(self, value: bool):
        value = bool(value)
        if self._visible == value:
            return

        self._visible = value

        for sprite_list in self.sprite_lists:
            sprite_list._update_color(self)

    @property
    def rgb(self) -> tuple[int, int, int]:
        """Get or set only the sprite's RGB color components.

        If a 4-color RGBA tuple is passed:

        * The new color's alpha value will be ignored
        * The old alpha value will be preserved
        """
        return self._color[:3]

    @rgb.setter
    def rgb(self, color: RGBOrA255):
        # Fast validation of size by unpacking channel values
        try:
            r, g, b, *_a = color
            if len(_a) > 1:  # Alpha's only used to validate here
                raise ValueError()

        except ValueError:  # It's always a length issue
            raise ValueError(
                f"{self.__class__.__name__},rgb takes 3 or 4 channel"
                f" colors, but got {len(color)} channels"
            )

        # Unpack to avoid index / . overhead & prep for repack
        current_r, current_b, current_g, a = self._color

        # Do nothing if equivalent to current color
        if current_r == r and current_g == g and current_b == b:
            return

        # Preserve the current alpha value & update sprite lists
        self._color = Color(r, g, b, a)
        for sprite_list in self.sprite_lists:
            sprite_list._update_color(self)

    @property
    def color(self) -> Color:
        """
        Get or set the RGBA multiply color for the sprite.

        When setting the color, it may be specified as any of the following:

        * an RGBA :py:class:`tuple` with each channel value between 0 and 255
        * an instance of :py:class:`~arcade.types.Color`
        * an RGB :py:class:`tuple`, in which case the color will be treated as opaque

        Example usage::

            >>> print(sprite.color)
            Color(255, 255, 255, 255)

            >>> sprite.color = arcade.color.RED

            >>> sprite.color = 255, 0, 0

            >>> sprite.color = 255, 0, 0, 128

        """
        return self._color

    @color.setter
    def color(self, color: RGBOrA255):
        if color == self._color:
            return

        r, g, b, *_a = color

        if _a:
            if len(_a) > 1:
                raise ValueError(f"iterable must unpack to 3 or 4 values not {len(color)}")
            a = _a[0]
        else:
            a = self._color.a

        # We don't handle alpha and .visible interactions here
        # because it's implemented in SpriteList._update_color
        self._color = Color(r, g, b, a)

        for sprite_list in self.sprite_lists:
            sprite_list._update_color(self)

    @property
    def alpha(self) -> int:
        """
        Get or set the alpha value of the sprite.

        Will be clamped to the range 0-255.
        """
        return self._color[3]

    @alpha.setter
    def alpha(self, alpha: int):
        self._color = Color(
            self._color[0], self._color[1], self._color[2], max(0, min(255, int(alpha)))
        )

        for sprite_list in self.sprite_lists:
            sprite_list._update_color(self)

    @property
    def texture(self) -> Texture:
        """
        Get or set the visible texture for this sprite
        This property can be changed over time to animate a sprite.

        Note that this doesn't change the hit box of the sprite.
        """
        return self._texture

    @texture.setter
    def texture(self, texture: Texture):
        if texture == self._texture:
            return

        if __debug__ and not isinstance(texture, Texture):
            raise TypeError(
                f"The 'texture' parameter must be an instance of arcade.Texture,"
                f" but is an instance of '{type(texture)}'."
            )

        self._texture = texture
        self._width = texture.width * self._scale[0]
        self._height = texture.height * self._scale[1]
        self.update_spatial_hash()
        for sprite_list in self.sprite_lists:
            sprite_list._update_texture(self)

    # ---- Update methods ----

    def update(self, delta_time: float = 1 / 60, *args, **kwargs) -> None:
        """
        Generic update method. It can be called manually
        or by the SpriteList's update method.

        Args:
            delta_time: Time since last update in seconds
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments
        """
        pass

    def update_animation(self, delta_time: float = 1 / 60, *args, **kwargs) -> None:
        """
        Generic update animation method. Usually involves changing
        the active texture on the sprite.

        This can be called manually or by the SpriteList's update_animation method.

        Args:
            delta_time: Time since last update in seconds
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments
        """
        pass

    # --- Scale methods -----

    def add_scale(self, factor: AsFloat) -> None:
        """Add to the sprite's scale by the factor.
        This adds the factor to both the x and y scale values.

        Args:
            factor: The factor to add to the sprite's scale.
        """
        self._scale = self._scale[0] + factor, self._scale[1] + factor
        self._hit_box.scale = self._scale
        tex_width, tex_height = self._texture.size
        self._width = tex_width * self._scale[0]
        self._height = tex_height * self._scale[1]

        self.update_spatial_hash()
        for sprite_list in self.sprite_lists:
            sprite_list._update_size(self)

    def multiply_scale(self, factor: AsFloat) -> None:
        """multiply the sprite's scale by the factor.
        This multiplies both the x and y scale values by the factor.

        Args:
            factor: The factor to scale up the sprite by.
        """

        self._scale = self._scale[0] * factor, self._scale[1] * factor
        self._hit_box.scale = self._scale
        tex_width, tex_height = self._texture.size
        self._width = tex_width * factor
        self._height = tex_height * factor

        self.update_spatial_hash()
        for sprite_list in self.sprite_lists:
            sprite_list._update_size(self)

    def rescale_relative_to_point(self, point: Point2, scale_by: AsFloat | Point2) -> None:
        """Rescale the sprite and its distance from the passed point.

        This function does two things:

        1. Multiply both values in the sprite's :py:attr:`.scale`
           value by the values in ``scale_by``:

           * If ``scale_by`` is an :py:class:`int` or :py:class:`float`,
             use it for both the x and y axes
           * If ``scale_by`` is a tuple-like object which unpacks to
             two numbers, then use
           * Otherwise, raise an exception

        2. Scale the distance between the sprite and ``point`` by
           ``factor``.

        .. note:: If ``point`` equals the sprite's :py:attr:`.position`
                  the distance will be zero and the sprite won't move.

        Args:
            point:
                The point to scale relative to.
            scale_by:
                A multiplier for both the sprite scale and its distance
                from the point. Note that although factor may be negative,
                it may have unexpected effects.

        """
        # abort if the multiplier wouldn't do anything
        if isinstance(scale_by, float | int):
            if scale_by == 1.0:
                return
            factor_x = scale_by
            factor_y = scale_by
        else:
            try:
                factor_x, factor_y = scale_by
                if factor_x == 1.0 and factor_y == 1.0:
                    return
            except ValueError:
                raise ValueError(
                    "factor must be a float, int, or tuple-like "
                    "which unpacks as two float-like values"
                )
            except TypeError:
                raise TypeError(
                    "factor must be a float, int, or tuple-like unpacks as two float-like values"
                )

        # set the scale and, if this sprite has a texture, the size data
        old_scale_x, old_scale_y = self._scale
        new_scale_x = old_scale_x * factor_x
        new_scale_y = old_scale_y * factor_y
        self._scale = new_scale_x, new_scale_y

        tex_width, tex_height = self._texture.size
        self._width = tex_width * new_scale_x
        self._height = tex_height * new_scale_y

        # If the scaling point is the sprite's center, it doesn't move
        old_position = self._position
        position_changed = point != old_position  # Stored to use below

        # be lazy about math; only do it if we have to
        if position_changed:
            point_x, point_y = point
            old_x, old_y = old_position
            self.position = (
                (old_x - point_x) * factor_x + point_x,
                (old_y - point_y) * factor_y + point_y,
            )

        # rebuild all spatial metadata
        self.update_spatial_hash()
        for sprite_list in self.sprite_lists:
            sprite_list._update_size(self)
            if position_changed:
                sprite_list._update_position(self)

    @warning(warning_type=ReplacementWarning, new_name="rescale_relative_to_point")
    def rescale_xy_relative_to_point(self, point: Point, factors_xy: Iterable[float]) -> None:
        """Rescale the sprite and its distance from the passed point.

        .. deprecated:: 3.0
           Use :py:meth:`.rescale_relative_to_point` instead.

           This was added during the 3.0 development cycle before scale was
           made into a vector quantity.

        This method can scale by different amounts on each axis. To
        scale along only one axis, set the other axis to ``1.0`` in
        ``factors_xy``.

        Internally, this function does the following:

        1. Multiply the x & y of the sprite's :py:attr:`~scale_xy`
           attribute by the corresponding part from ``factors_xy``.
        2. Scale the x & y of the difference between the sprite's
           position and ``point`` by the corresponding component from
           ``factors_xy``.

        If ``point`` equals the sprite's :py:attr:`~position`,
        the distance will be zero and the sprite will not move.

        Args:
            point:
                The reference point for rescaling.
            factors_xy:
                A 2-length iterable containing x and y
                multipliers for ``scale`` & distance to ``point``.
        """
        self.rescale_relative_to_point(point, factors_xy)  # type: ignore

    # ---- Utility Methods ----

    @property
    def hit_box(self) -> HitBox:
        """The hit box for this sprite."""
        return self._hit_box

    def update_spatial_hash(self) -> None:
        """Update the sprites location in the spatial hash if present."""
        for sprite_list in self.sprite_lists:
            if sprite_list.spatial_hash is not None:
                sprite_list.spatial_hash.move(self)

    def register_sprite_list(self: SpriteType, new_list: SpriteList[SpriteType]) -> None:
        """
        Register this sprite as belonging to a list.

        We will automatically remove ourselves from the list when kill() is called.
        """
        self.sprite_lists.append(new_list)

    def _unregister_sprite_list(self: SpriteType, new_list: SpriteList[SpriteType]) -> None:
        """Unregister this sprite as belonging to a list."""
        self.sprite_lists.remove(new_list)

    def remove_from_sprite_lists(self) -> None:
        """Remove the sprite from all sprite lists."""
        while len(self.sprite_lists) > 0:
            self.sprite_lists[0].remove(self)

    # ----- Drawing Methods -----

    def draw_hit_box(self, color: RGBOrA255 = BLACK, line_thickness: float = 2.0) -> None:
        """
        Draw a sprite's hit-box. This is useful for debugging.

        Args:
            color:
                Color of box
            line_thickness:
                How thick the box should be
        """
        converted_color = Color.from_iterable(color)
        points: Point2List = self.hit_box.get_adjusted_points()
        # NOTE: This is a COPY operation. We don't want to modify the points.
        points = tuple(points) + tuple(points[:-1])
        arcade.draw_line_strip(points, color=converted_color, line_width=line_thickness)

    # ---- Shortcut Methods ----

    def kill(self) -> None:
        """
        Alias of ``remove_from_sprite_lists()``.
        """
        self.remove_from_sprite_lists()

    def collides_with_point(self, point: Point2) -> bool:
        """
        Check if point is within the current sprite.

        Args:
            point: Point to check.
        Returns:
            ``True`` if the point is contained within the sprite's boundary.
        """
        from arcade.geometry import is_point_in_polygon

        x, y = point
        return is_point_in_polygon(x, y, self.hit_box.get_adjusted_points())

    def collides_with_sprite(self, other: BasicSprite) -> bool:
        """Will check if a sprite is overlapping (colliding) another Sprite.

        Args:
            other: the other sprite to check against.
        Returns:
            ``True`` or ``False``, whether or not they are overlapping.
        """
        from arcade import check_for_collision

        return check_for_collision(self, other)

    def collides_with_list(self, sprite_list: SpriteList[SpriteType]) -> list[SpriteType]:
        """Check if current sprite is overlapping with any other sprite in a list

        Args:
            sprite_list: SpriteList to check against
        Returns:
            List of all overlapping Sprites from the original SpriteList
        """
        from arcade import check_for_collision_with_list

        return check_for_collision_with_list(self, sprite_list)



from array import array

import arcade
from arcade.gl import Buffer, BufferDescription, Geometry, Program
from arcade.math import Vec2
from arcade.texture_atlas.base import TextureAtlasBase
from arcade.types import Rect


class NinePatchTexture:
    """Keeps borders & corners at constant widths while stretching the middle.

    It can be used with new or existing :py:class:`~arcade.gui.UIWidget`
    subclasses wherever an ordinary :py:class:`arcade.Texture` is
    supported. This is useful for GUI elements which must grow or shrink
    while keeping their border decorations constant, such as dialog boxes
    or text boxes.

    The diagram below explains the stretching behavior of this class:

    * Numbered regions with arrows (``<--->``) stretch along the
      direction(s) of any arrows present
    * Bars (``|---|``) mark the distances specified by the border
      parameters (``left``, etc)

    .. code-block::
        :caption: Stretch Axes & Border Parameters

            left                        right
            |------|                 |------|
                                               top
            +------+-----------------+------+  ---
            | (1)  | (2)             | (3)  |   |
            |      | <-------------> |      |   |
            +------+-----------------+------+  ---
            | (4)  | (5)    ^        | (6)  |
            |  ^   |        |        |   ^  |
            |  |   |        |        |   |  |
            |  |   | <------+------> |   |  |
            |  |   |        |        |   |  |
            |  |   |        |        |   |  |
            |  v   |        v        |   v  |
            +------+-----------------+------+  ---
            | (7)  | (8)             | (9)  |   |
            |      | <-------------> |      |   |
            +------+-----------------+------+  ---
                                              bottom

    As the texture is stretched, the numbered slices of the texture behave
    as follows:

    * Areas ``(1)``, ``(3)``, ``(7)`` and ``(9)`` never stretch.
    * Area ``(5)`` stretches both horizontally and vertically.
    * Areas ``(2)`` and ``(8)`` only stretch horizontally.
    * Areas ``(4)`` and ``(6)`` only stretch vertically.

    Args:
        left: The width of the left border of the 9-patch (in pixels)
        right: The width of the right border of the 9-patch (in pixels)
        bottom: The height of the bottom border of the 9-patch (in
            pixels)
        top: The height of the top border of the 9-patch (in pixels)
        texture: The raw texture to use for the 9-patch
        atlas: Specify an atlas other than Arcade's default texture
            atlas
    """

    def __init__(
        self,
        left: int,
        right: int,
        bottom: int,
        top: int,
        texture: arcade.Texture,
        *,
        atlas: TextureAtlasBase | None = None,
    ):
        self._initialized = False
        self._texture = texture
        self._custom_atlas = atlas
        self._geometry_cache: tuple[int, int, int, int, Rect] | None = None

        # pixel texture co-ordinate start and end of central box.
        self._left = left
        self._right = right
        self._bottom = bottom
        self._top = top

        self._check_sizes()

        # Created in _init_deferred
        self._buffer: Buffer
        self._program: Program
        self._geometry: Geometry
        self._ctx: arcade.ArcadeContext
        self._atlas: TextureAtlasBase
        try:
            self._init_deferred()
        except Exception:
            pass

    def initialize(self) -> None:
        """
        Manually initialize the NinePatchTexture if it was lazy loaded.
        This has no effect if the NinePatchTexture was already initialized.
        """
        if not self._initialized:
            self._init_deferred()

    @property
    def ctx(self) -> arcade.ArcadeContext:
        """The OpenGL context."""
        if not self._initialized:
            raise RuntimeError("The NinePatchTexture has not been initialized")
        return self._ctx

    @property
    def texture(self) -> arcade.Texture:
        """Get or set the texture."""
        return self._texture

    @texture.setter
    def texture(self, texture: arcade.Texture):
        self._texture = texture
        self._add_to_atlas(texture)

    @property
    def program(self) -> Program:
        """Get or set the shader program.

        Returns the default shader if no other shader is assigned.
        """
        if not self._initialized:
            raise RuntimeError("The NinePatchTexture has not been initialized")

        return self._program

    @program.setter
    def program(self, program: Program):
        if not self._initialized:
            raise RuntimeError("The NinePatchTexture has not been initialized")

        self._program = program

    def _add_to_atlas(self, texture: arcade.Texture):
        """Internal method for setting the texture.

        It ensures the texture is added to the global atlas.
        """
        if not self._atlas.has_texture(texture):
            self._atlas.add(texture)

    @property
    def left(self) -> int:
        """Get or set the left border of the 9-patch."""
        return self._left

    @left.setter
    def left(self, left: int):
        self._left = left

    @property
    def right(self) -> int:
        """Get or set the right border of the 9-patch."""
        return self._right

    @right.setter
    def right(self, right: int):
        self._right = right

    @property
    def bottom(self) -> int:
        """Get or set the bottom border of the 9-patch."""
        return self._bottom

    @bottom.setter
    def bottom(self, bottom: int):
        self._bottom = bottom

    @property
    def top(self) -> int:
        """Get or set the top border of the 9-patch."""
        return self._top

    @top.setter
    def top(self, top: int):
        self._top = top

    @property
    def size(self) -> tuple[int, int]:
        """The size of texture as a width, height tuple in pixels."""
        return self.texture.size

    @property
    def width(self) -> int:
        """The width of the texture in pixels."""
        return self.texture.width

    @property
    def height(self) -> int:
        """The height of the texture in pixels."""
        return self.texture.height

    def draw_rect(
        self,
        *,
        rect: arcade.types.Rect,
        pixelated: bool = True,
        blend: bool = True,
        **kwargs,
    ):
        """Draw the 9-patch texture with a specific size.

        Warning:
            This method assumes the passed dimensions are proper!

            Unexpected behavior may occur if you specify a size
            smaller than the total size of the border areas.

        Args:
            rect: Rectangle to draw the 9-patch texture in
            pixelated: Whether to draw with nearest neighbor
                interpolation
        """
        if not self._initialized:
            self._init_deferred()

        self._create_geometry(rect)

        if blend:
            self._ctx.enable_only(self._ctx.BLEND)
        else:
            self._ctx.disable(self._ctx.BLEND)

        if pixelated:
            self._atlas.texture.filter = self._ctx.NEAREST, self._ctx.NEAREST
        else:
            self._atlas.texture.filter = self._ctx.LINEAR, self._ctx.LINEAR

        self._atlas.use_uv_texture(0)
        self._atlas.texture.use(1)

        self._geometry.render(self._program)

        if blend:
            self._ctx.disable(self._ctx.BLEND)

    def _check_sizes(self):
        """Raise a ValueError if any dimension is invalid."""
        # Sanity check values
        if self._left < 0:
            raise ValueError("Left border must be a positive integer")
        if self._right < 0:
            raise ValueError("Right border must be a positive integer")
        if self._bottom < 0:
            raise ValueError("Bottom border must be a positive integer")
        if self._top < 0:
            raise ValueError("Top border must be a positive integer")

        # Sanity check texture size
        if self._left + self._right > self._texture.width:
            raise ValueError("Left and right border must be smaller than texture width")
        if self._bottom + self._top > self._texture.height:
            raise ValueError("Bottom and top border must be smaller than texture height")

    def _init_deferred(self):
        """Deferred initialization when lazy loaded"""
        self._ctx = arcade.get_window().ctx
        # TODO: Cache in context?
        self._program = self._ctx.load_program(
            vertex_shader=":system:shaders/gui/nine_patch_vs.glsl",
            fragment_shader=":system:shaders/gui/nine_patch_fs.glsl",
        )
        # Configure texture channels
        self._program.set_uniform_safe("uv_texture", 0)
        self._program["sprite_texture"] = 1

        # 4 byte floats * 4 floats * 4 vertices * 9 patches
        self._buffer = self._ctx.buffer(reserve=576)
        # fmt: off
        self._ibo = self._ctx.buffer(
            data=array("i",
                [
                    # Triangulate the patches
                    # First rot
                    0, 1, 2,
                    3, 1, 2,

                    4, 5, 6,
                    7, 5, 6,

                    8, 9, 10,
                    11, 9, 10,

                    # Middle row
                    12, 13, 14,
                    15, 13, 14,

                    16, 17, 18,
                    19, 17, 18,

                    20, 21, 22,
                    23, 21, 22,

                    # Bottom row
                    24, 25, 26,
                    27, 25, 26,

                    28, 29, 30,
                    31, 29, 30,

                    32, 33, 34,
                    35, 33, 34,
                ]
            ),
        )
        # fmt: on
        self._geometry = self._ctx.geometry(
            content=[BufferDescription(self._buffer, "2f 2f", ["in_position", "in_uv"])],
            index_buffer=self._ibo,
            mode=self._ctx.TRIANGLES,
            index_element_size=4,
        )

        # References for the texture
        self._atlas = self._custom_atlas or self._ctx.default_atlas
        self._add_to_atlas(self.texture)

        # NOTE: Important to create geometry after the texture is added to the atlas
        # self._create_geometry(LBWH(0, 0, self.width, self.height))
        self._initialized = True

    def _create_geometry(self, rect: Rect):
        """Create vertices for the 9-patch texture."""
        # NOTE: This was ported from glsl geometry shader to python
        # Simulate old uniforms
        cache_key = (self._left, self._right, self._bottom, self._top, rect)
        if cache_key == self._geometry_cache:
            return
        self._geometry_cache = cache_key

        position = rect.bottom_left
        start = Vec2(self._left, self._bottom)
        end = Vec2(self.width - self._right, self.height - self._top)
        size = rect.size
        t_size = Vec2(*self._texture.size)
        atlas_size = Vec2(*self._atlas.size)

        # Patch points starting from upper left row by row
        p1 = position + Vec2(0.0, size.y)
        p2 = position + Vec2(start.x, size.y)
        p3 = position + Vec2(size.x - (t_size.x - end.x), size.y)
        p4 = position + Vec2(size.x, size.y)

        y = size.y - (t_size.y - end.y)
        p5 = position + Vec2(0.0, y)
        p6 = position + Vec2(start.x, y)
        p7 = position + Vec2(size.x - (t_size.x - end.x), y)
        p8 = position + Vec2(size.x, y)

        p9 = position + Vec2(0.0, start.y)
        p10 = position + Vec2(start.x, start.y)
        p11 = position + Vec2(size.x - (t_size.x - end.x), start.y)
        p12 = position + Vec2(size.x, start.y)

        p13 = position + Vec2(0.0, 0.0)
        p14 = position + Vec2(start.x, 0.0)
        p15 = position + Vec2(size.x - (t_size.x - end.x), 0.0)
        p16 = position + Vec2(size.x, 0.0)

        # <AtlasRegion
        #     x=1 y=1
        #     width=100 height=100
        #     uvs=(
        #         0.001953125, 0.001953125,
        #         0.197265625, 0.001953125,
        #         0.001953125, 0.197265625,
        #         0.197265625, 0.197265625,
        #     )
        # Get texture coordinates
        # vec2 uv0, uv1, uv2, uv3
        region = self._atlas.get_texture_region_info(self._texture.atlas_name)
        tex_coords = region.texture_coordinates
        uv0 = Vec2(tex_coords[0], tex_coords[1])
        uv1 = Vec2(tex_coords[2], tex_coords[3])
        uv2 = Vec2(tex_coords[4], tex_coords[5])
        uv3 = Vec2(tex_coords[6], tex_coords[7])

        # Local corner offsets in pixels
        left = start.x
        right = t_size.x - end.x
        top = t_size.y - end.y
        bottom = start.y

        # UV offsets to the inner rectangle in the patch
        # This is the global texture coordinate offset in the entire atlas
        c1 = Vec2(left, top) / atlas_size  # Upper left corner
        c2 = Vec2(right, top) / atlas_size  # Upper right corner
        c3 = Vec2(left, bottom) / atlas_size  # Lower left corner
        c4 = Vec2(right, bottom) / atlas_size  # Lower right corner

        # Texture coordinates for all the points in the patch
        t1 = uv0
        t2 = uv0 + Vec2(c1.x, 0.0)
        t3 = uv1 - Vec2(c2.x, 0.0)
        t4 = uv1

        t5 = uv0 + Vec2(0.0, c1.y)
        t6 = uv0 + c1
        t7 = uv1 + Vec2(-c2.x, c2.y)
        t8 = uv1 + Vec2(0.0, c2.y)

        t9 = uv2 - Vec2(0.0, c3.y)
        t10 = uv2 + Vec2(c3.x, -c3.y)
        t11 = uv3 - c4
        t12 = uv3 - Vec2(0.0, c4.y)

        t13 = uv2
        t14 = uv2 + Vec2(c3.x, 0.0)
        t15 = uv3 - Vec2(c4.x, 0.0)
        t16 = uv3

        # fmt: off
        primitives = [
            # First row - two fixed corners + stretchy middle
            # Upper left corner. Fixed size.
            p1, t1,
            p5, t5,
            p2, t2,
            p6, t6,
            # Upper middle part stretches on x axis
            p2, t2,
            p6, t6,
            p3, t3,
            p7, t7,
            # Upper right corner. Fixed size
            p3, t3,
            p7, t7,
            p4, t4,
            p8, t8,

            #  Middle row: Two stretchy sides + stretchy middle
            # left border sketching on y axis
            p5, t5,
            p9, t9,
            p6, t6,
            p10, t10,
            # Center stretchy area
            p6, t6,
            p10, t10,
            p7, t7,
            p11, t11,
            # Right border. Stenches on y axis
            p7, t7,
            p11, t11,
            p8, t8,
            p12, t12,

            # Bottom row: two fixed corners + stretchy middle
            # Lower left corner. Fixed size.
            p9, t9,
            p13, t13,
            p10, t10,
            p14, t14,
            # Lower middle part stretches on x axis
            p10, t10,
            p14, t14,
            p11, t11,
            p15, t15,
            # Lower right corner. Fixed size
            p11, t11,
            p15, t15,
            p12, t12,
            p16, t16,
        ]
        # fmt: on

        data = array("f", [coord for point in primitives for coord in point])
        self._buffer.write(data.tobytes())



"""
Perspective example using the PerspectiveProjectionCamera

This is definitely in the advanced section, but it can be
a useful tool to learn. Sometimes we want perspective
projection for things like backgrounds. This can be
done very efficiently with shaders.

In this example we render content into a framebuffer /
virtual screen and map that on a texture we can rotate
in 3D.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.perspective
"""

from array import array

import arcade
from arcade.gl import BufferDescription


class GameView(arcade.View):

    def __init__(self):
        super().__init__()
        # Simple texture shader for the plane.
        # It support projection and model matrix
        # and a scroll value for texture coordinates
        self.program = self.window.ctx.program(
            vertex_shader="""
            #version 330

            uniform WindowBlock {
                mat4 projection;
                mat4 model;
            } window;

            in vec3 in_pos;
            in vec2 in_uv;

            out vec2 uv;

            void main() {
                gl_Position = window.projection * window.model * vec4(in_pos, 1.0);
                uv = in_uv;
            }
            """,
            fragment_shader="""
            #version 330

            uniform sampler2D layer;
            uniform vec2 scroll;

            in vec2 uv;
            out vec4 fragColor;

            void main() {
                fragColor = texture(layer, uv + scroll);
            }
            """,
        )

        # Configure and create the perspective projector
        self.perspective_data = arcade.camera.PerspectiveProjectionData(
            self.window.aspect_ratio, # The ratio between window width and height
            75, # The angle  between things at the top of the screen, and the bottom
            0.1, # Anything within 0.1 units of the camera won't be visible
            100.0 # Anything past 100.0 units of the camera won't be visible
        )
        self.projector = arcade.camera.PerspectiveProjector()

        # Framebuffer / virtual screen to render the contents into
        self.fbo = self.window.ctx.framebuffer(
            color_attachments=self.window.ctx.texture(size=(1024, 1024))
        )

        # Set up the geometry buffer for the plane.
        # This is four points with texture coordinates
        # creating a rectangle
        buffer = self.window.ctx.buffer(
            data=array(
                'f',
                [
                    # x  y   z  u  v
                    -1,  1, 0, 0, 1,  # Top Left
                    -1, -1, 0, 0, 0,  # Bottom Left
                     1,  1, 0, 1, 1,  # Top Right
                     1, -1, 0, 1, 0,  # Bottom right
                ]
            )
        )
        # Make this into a geometry object we can draw-
        # Here we describe the contents of the buffer so the shader can understand it
        self.geometry = self.window.ctx.geometry(
            content=[BufferDescription(buffer, "3f 2f", ("in_pos", "in_uv"))],
            mode=arcade.gl.TRIANGLE_STRIP,
        )

        # Create some sprites
        self.spritelist = arcade.SpriteList()
        for y in range(8):
            for x in range(8):
                self.spritelist.append(
                    arcade.Sprite(
                        ":resources:images/tiles/boxCrate_double.png",
                        center_x=64 + x * 128,
                        center_y=64 + y * 128,
                    )
                )

        # Create a 2D camera for rendering to the fbo
        # by setting the camera's render target it will automatically
        # size and position itself correctly
        self.offscreen_cam = arcade.Camera2D(
            render_target=self.fbo
        )

    def on_update(self, delta_time: float):
        # Rotate the perspective camera around the plane
        view_data = self.projector.view
        view_data.position = arcade.math.quaternion_rotation(
            (1.0, 0.0, 0.0), (0, 0, 3), 180 * self.window.time
        )
        view_data.forward, view_data.up = arcade.camera.grips.look_at(view_data, (0.0, 0.0, 0.0))
        # print(view_data)


    def on_draw(self):
        # Every frame we can update the offscreen texture if needed
        self.draw_offscreen()
        # Clear the window
        self.clear()

        with self.projector.activate():
            # Bind the texture containing the offscreen data to channel 0
            self.fbo.color_attachments[0].use(unit=0)

            # Scroll the texture coordinates
            self.program["scroll"] = 0, -self.window.time / 5

            # Draw the plane
            self.geometry.render(self.program)

    def draw_offscreen(self):
        """Render into the texture mapped """
        # Activate the offscreen cam, this also activates it's render target
        with self.offscreen_cam.activate():
            self.fbo.clear()
            self.offscreen_cam.use()
            self.spritelist.draw()

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.perspective_data.aspect = height / width


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(1280, 720, "Perspective Example", resizable=True)

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



__all__ = (
    "Clock",
    "FixedClock",
    "GLOBAL_CLOCK",
    "GLOBAL_FIXED_CLOCK",
)


class Clock:
    """
    A time keeping class which provides a method for easily tracking the
    elapsed time, delta_time, and number of ticks.

    Arcade provides a global clock which is automatically ticked by the window.

    *Coming post 3.0:*
    you can add 'sub-clocks' to Arcade's top level clock which will tick at the
    same time, and have cumulative tick_speeds. This allows you to slow down
    only certain elements rather than everything.

    Args:
        initial_elapsed: The amount of time the clock should assume has
            already occurred. Defaults to 0.0
        initial_tick: The number of ticks the clock should assume has already
            occurred. Defaults to 0.
        tick_speed: A multiplier on how the 'speed' of time.
            i.e. a value of 0.5 means time elapsed half as fast for this clock. Defaults to 1.0.
    """

    def __init__(
        self, initial_elapsed: float = 0.0, initial_tick: int = 0, tick_speed: float = 1.0
    ):
        self._elapsed_time: float = initial_elapsed
        self._tick: int = initial_tick
        self._tick_delta_time: float = 0.0
        self._tick_speed: float = tick_speed

        self._max_deltatime: float | None = None

    def tick(self, delta_time: float):
        """
        Update the clock with the time that has passed since the last tick.

        Args:
            delta_time: The amount of time that has passed since the last tick.
        """
        if self._max_deltatime is not None:
            delta_time = min(self._max_deltatime, delta_time)

        self._tick_delta_time = delta_time * self._tick_speed
        self._elapsed_time += self._tick_delta_time
        self._tick += 1

    def set_tick_speed(self, new_tick_speed: float):
        """
        Set the speed of time for this clock.

        Args:
            new_tick_speed: A multiplier on the 'speed' of time.
                i.e. a value of 0.5 means time elapsed half as fast for this clock.
        """
        self._tick_speed = new_tick_speed

    def set_max_deltatime(self, max_deltatime: float | None = None):
        """
        Set the maximum deltatime that the clock will allow. If a large dt is passed into
        the clock's tick method it will be clamped. This will desync the game's time with the
        real world elapsed time, but can help protect against lag-spikes, debugger pauses, and
        other pauses to the event loop. This impacts the 'raw' dt so it does not take the clock's
        tick speed into account

        Args:
            max_deltatime: The maximum number of seconds that a clock can have as it's deltatime.
                           If set to None the clock has no limit. Defaults to None.
        """
        self._max_deltatime = max_deltatime

    def time_since(self, time: float) -> float:
        """
        Calculate the amount of time that has passed since the given time.

        Args:
            time: The time to compare against.
        """
        return self._elapsed_time - time

    def ticks_since(self, tick: int) -> int:
        """
        Calculate the number of ticks that have occurred since the given tick.

        Args:
            tick: The tick to compare against.
        """
        return self._tick - tick

    @property
    def max_deltatime(self) -> float | None:
        """
        The maximum deltatime that the clock will allow. If a large dt is passed into
        """
        return self._max_deltatime

    @property
    def time(self) -> float:
        """The total number of seconds that have elapsed for this clock"""
        return self._elapsed_time

    @property
    def t(self) -> float:
        """Alias to Clock.time"""
        return self._elapsed_time

    @property
    def delta_time(self) -> float:
        """The amount of time that elapsed during the last tick"""
        return self._tick_delta_time

    @property
    def dt(self) -> float:
        """Alias to Clock.delta_time"""
        return self.delta_time

    @property
    def speed(self) -> float:
        """
        A modifier on the delta time that elapsed each tick.

        Decreasing the speed will 'slow down' time for this clock
        Immutable in 3.0
        """
        return self._tick_speed

    @property
    def ticks(self) -> int:
        """The number of ticks that have occurred for this clock."""
        return self._tick

    @property
    def tick_count(self) -> int:
        """Alias to Clock.ticks"""
        return self._tick


class FixedClock(Clock):
    """
    A fixed clock which expects its delta_time to stay constant. If it doesn't it
    will throw an error.

    Arcade provides a global fixed clock which is automatically ticked every update

    Args:
        sibling: The unfixed clock which this clock will sync with.
        fixed_tick_rate: The fixed number of seconds that pass
            for this clock every tick. Defaults to ``1.0 / 60.0``.
    """

    def __init__(self, sibling: Clock, fixed_tick_rate: float = 1.0 / 60.0):
        self._sibling_clock: Clock = sibling
        self._fixed_rate: float = fixed_tick_rate
        super().__init__()

    def set_tick_speed(self, new_tick_speed: float):
        """
        Set the speed of time for this clock.

        Args:
            new_tick_speed: A multiplier on the 'speed' of time.
                i.e. a value of 0.5 means time elapsed half as fast for this clock
        """
        raise ValueError(
            "It is not safe to change the tick speed of a fixed clock post initialization."
        )

    def tick(self, delta_time: float):
        """
        Update the clock with the time that has passed since the last tick.

        Args:
            delta_time: The amount of time that has passed since the last tick.
        """
        if delta_time != self._fixed_rate:
            raise ValueError(
                f"the delta_time {delta_time}, "
                f"does not match the fixed clock's required delta_time {self._fixed_rate}"
            )
        super().tick(self._fixed_rate)

    @property
    def rate(self) -> float:
        """The fixed number of seconds that pass for this clock every tick."""
        return self._fixed_rate

    @property
    def accumulated(self) -> float:
        """The total number of seconds that have elapsed for this clock"""
        return self._sibling_clock.time - self._elapsed_time

    @property
    def fraction(self) -> float:
        """The fraction of a fixed tick that has passed since the last tick."""
        return self.accumulated / self._fixed_rate


GLOBAL_CLOCK = Clock()
GLOBAL_FIXED_CLOCK = FixedClock(sibling=GLOBAL_CLOCK)


def _setup_clock(initial_elapsed: float = 0.0, initial_tick: int = 0, tick_speed: float = 1.0):
    """
    Private method used by the Arcade window to setup the global clock post initialization.

    Args:
        initial_elapsed: The amount of time the clock should assume
            has already occurred. Defaults to 0.0
        initial_tick: The number of ticks the clock should assume has
            already occurred. Defaults to 0.
        tick_speed: A multiplier on the 'speed' of time.
            i.e. a value of 0.5 means time elapsed half as fast for this clock.
            Defaults to 1.0.
    """
    GLOBAL_CLOCK._elapsed_time = initial_elapsed  # noqa: SLF001
    GLOBAL_CLOCK._tick = initial_tick  # noqa: SLF001
    GLOBAL_CLOCK._tick_speed = tick_speed  # noqa: SLF001


def _setup_fixed_clock(fixed_tick_rate: float = 1.0 / 60.0):
    """
    Private method used by the Arcade window to setup the global fixed clock
    post initialization.

    Args:
        fixed_tick_rate: The fixed number of seconds that pass
            for this clock every tick. Defaults to 1.0 / 60.0
    """
    GLOBAL_FIXED_CLOCK._elapsed_time = GLOBAL_CLOCK.time  # noqa: SLF001
    GLOBAL_FIXED_CLOCK._tick = GLOBAL_CLOCK.tick_count  # noqa: SLF001
    GLOBAL_FIXED_CLOCK._fixed_rate = fixed_tick_rate  # noqa: SLF001



from pathlib import Path
from collections.abc import Sequence
from arcade.exceptions import warning, ReplacementWarning

#: The absolute path to this directory
RESOURCE_DIR = Path(__file__).parent.resolve()

# The "system" resources common to Arcade
SYSTEM_PATH = RESOURCE_DIR / "system"
FONTS_PATH = SYSTEM_PATH / "fonts"
TTF_PATH = FONTS_PATH / "ttf"

# Basic resources in the :assets: handle
ASSET_PATH = RESOURCE_DIR / "assets"


handles: dict[str, list[Path]] = {
    "resources": [SYSTEM_PATH, ASSET_PATH],
    "assets": [ASSET_PATH],
    "system": [SYSTEM_PATH],
}

__all__ = [
    "resolve_resource_path",
    "resolve",
    "add_resource_handle",
    "get_resource_handle_paths",
]


@warning(warning_type=ReplacementWarning, new_name="resolve")
def resolve_resource_path(path: str | Path) -> Path:
    """
    Attempts to resolve a path to a resource including resource handles.

    If the path is a string it tries to resolve it as a resource handle
    or convert it to a Path object.

    If the path is a Path object it will :py:meth:`~pathlib.Path.resolve` it
    unless it's not absolute and return it.

    Example::

        resolve(":resources:images/cards/cardBack_blue1.png")
        resolve(":my_handle:music/combat.wav")

    Args:
        path: A Path or string
    """
    return resolve(path)


def resolve(path: str | Path) -> Path:
    """
    Attempts to resolve a path to a resource including resource handles.

    If the path is a string it tries to resolve it as a resource handle
    or convert it to a Path object.

    If the path is a Path object it will :py:meth:`~pathlib.Path.resolve` it
    unless it's not absolute and return it.

    Example::

        resolve(":resources:images/cards/cardBack_blue1.png")
        resolve(":my_handle:music/combat.wav")

    Args:
        path: A Path or string
    """
    # Convert to a Path object and resolve resource handle
    if isinstance(path, str):
        path = path.strip()  # Allow for silly mistakes with extra spaces

        # If the path starts with a colon, it's a resource handle
        if path.startswith(":"):
            path = path[1:]
            try:
                handle, resource = path.split(":")
            except ValueError:
                raise ValueError(f"Invalid resource handle '{path}'")

            while resource.startswith("/") or resource.startswith("\\"):
                resource = resource[1:]

            # Iterate through the paths in reverse order to find the first
            # match. This allows for overriding of resources.
            paths = get_resource_handle_paths(handle)
            for handle_path in reversed(paths):
                path = handle_path / resource
                if path.exists():
                    break
            else:
                searched_paths = "\n".join(f"-> {p}" for p in reversed(paths))
                raise FileNotFoundError(
                    f"Cannot locate resource '{resource}' using handle "
                    f"'{handle}' in any of the following paths:\n"
                    f"{searched_paths}"
                )

            # Always convert into a Path object
            path = Path(handle_path / resource)
        else:
            path = Path(path)

    try:
        path = Path(path.resolve(strict=True))
    except AttributeError:
        # WARNING: This is due to an issue caused by Nuitka overriding strings
        #          into janky path object
        path = Path(path.absolute())

    # Always return absolute paths
    # Check for the existence of the file and provide useful feedback to
    # avoid deep stack trace into pathlib
    try:
        # If the path is absolute, just return it. We assume it's valid and resolved.
        if path.is_absolute():
            return path
        return path.resolve(strict=True)
    except FileNotFoundError:
        raise FileNotFoundError(f"Cannot locate resource : {path}")


def add_resource_handle(handle: str, path: str | Path) -> None:
    """
    Add a resource handle or path to an existing handle.

    A handle can point to multiple paths. If a resource is not found in
    the first path, it will look in the next path, and so on. The search
    is done in reverse order, so the last path added is searched first.

    Args:
        handle: The name of the handle
        path: The absolute path to a directory
    """
    if isinstance(path, str):
        path = Path(path)
    elif isinstance(path, Path):
        path = path
    else:
        raise TypeError("Path for resource handle must be a string or Path object")

    if not path.is_absolute():
        raise RuntimeError(
            f"Path for resource handle must be absolute, not relative ('{path}'). "
            "See https://docs.python.org/3/library/pathlib.html#pathlib.Path.resolve"
        )

    if not path.exists():
        raise FileNotFoundError(f"Directory '{path}' for handle '{handle}' does not exist")

    paths = handles.setdefault(handle, [])
    # Don't allow duplicate paths
    if path not in paths:
        paths.append(path)


def get_resource_handle_paths(handle: str) -> list[Path]:
    """
    Returns the paths for a resource handle.

    Args:
        handle: The name of the handle
    """
    try:
        return handles[handle]
    except KeyError:
        raise KeyError(f'Unknown resource handle "{handle}"')


def list_built_in_assets(
    *, name: str | None = None, extensions: Sequence[str] | None = None
) -> list[Path]:
    """
    List built in assets in arcade.

    This will traverse the assets directory returning a list of resources
    optionally filtered by name and file extensions.

    Example::

        # Get all assets
        list_built_in_assets()

        # Only get .png files
        list_built_in_assets(extensions=(".png",))

        # Get all card images
        list_built_in_assets(name="card", extensions=(".png", ".jpg"))

    Args:
        name: Include only assets that contain this string in the filename
        extensions: A tuple of file extensions to filter by
    Returns:
        A list of absolute paths to requested assets
    """
    all_paths = ASSET_PATH.glob("**/*")
    if extensions is None and name is None:
        return list(all_paths)

    if name:
        name = name.lower()

    filtered_paths: list[Path] = []
    for path in all_paths:
        if extensions and path.suffix not in extensions:
            continue
        if name and name not in path.name.lower():
            continue

        filtered_paths.append(path)

    return filtered_paths


def load_kenney_fonts() -> None:
    """Loads all the Kenney.nl fonts bundled with Arcade.

    .. tip:: This function is best for prototyping and experimenting!

         For best performance, you may want to switch to
         :py:class:`arcade.load_font` before release.

    Please see :ref:`resources-fonts-kenney` for previews and
    license information. The filename to load and ``font_name`` to use
    when drawing text are summarized below:

    .. might swap to this style for the resources listing once I figure out how to
    .. cleanly modify the file to use it.

    =========================================  =========================================================================
    ``font_name`` for :py:class:`arcade.Text`  :ref:`Resource handle <resource_handles>` for :py:func:`arcade.load_font`
    =========================================  =========================================================================
    ``"Kenney Blocks"``                        ``:resources:fonts/ttf/Kenney/Kenney_Blocks.ttf``
    ``"Kenney Future"``                        ``:resources:fonts/ttf/Kenney/Kenney_Future.ttf``
    ``"Kenney Future Narrow"``                 ``:resources:fonts/ttf/Kenney/Kenney_Future_Narrow.ttf``
    ``"Kenney High"``                          ``:resources:fonts/ttf/Kenney/Kenney_High.ttf``
    ``"Kenney High Square"``                   ``:resources:fonts/ttf/Kenney/Kenney_High_Square.ttf``
    ``"Kenney Mini"``                          ``:resources:fonts/ttf/Kenney/Kenney_Mini.ttf``
    ``"Kenney Mini Square"``                   ``:resources:fonts/ttf/Kenney/Kenney_Mini_Square.ttf``
    ``"Kenney Pixel"``                         ``:resources:fonts/ttf/Kenney/Kenney_Pixel.ttf``
    ``"Kenney Pixel Square"``                  ``:resources:fonts/ttf/Kenney/Kenney_Pixel_Square.ttf``
    ``"Kenney Rocket"``                        ``:resources:fonts/ttf/Kenney/Kenney_Rocket.ttf``
    ``"Kenney Rocket Square"``                 ``:resources:fonts/ttf/Kenney/Kenney_Rocket_Square.ttf``
    =========================================  =========================================================================

    """  # noqa: E501  # Silence ruff  # pending: better generation
    from arcade.text import load_font

    load_font(":system:fonts/ttf/Kenney/Kenney_Blocks.ttf")
    load_font(":system:fonts/ttf/Kenney/Kenney_Future.ttf")
    load_font(":system:fonts/ttf/Kenney/Kenney_Future_Narrow.ttf")
    load_font(":system:fonts/ttf/Kenney/Kenney_High.ttf")
    load_font(":system:fonts/ttf/Kenney/Kenney_High_Square.ttf")
    load_font(":system:fonts/ttf/Kenney/Kenney_Mini.ttf")
    load_font(":system:fonts/ttf/Kenney/Kenney_Mini_Square.ttf")
    load_font(":system:fonts/ttf/Kenney/Kenney_Pixel.ttf")
    load_font(":system:fonts/ttf/Kenney/Kenney_Pixel_Square.ttf")
    load_font(":system:fonts/ttf/Kenney/Kenney_Rocket.ttf")
    load_font(":system:fonts/ttf/Kenney/Kenney_Rocket_Square.ttf")


def load_liberation_fonts() -> None:
    """Loads all styles for generic Arial, Courier, and Times New Roman replacements.

    .. tip:: This function is best for prototyping and experimenting!

             For best performance, you may want to switch to
             :py:class:`arcade.load_font` before release.

    The Liberation fonts are proven, permissively-licensed fonts.[
    For previews and additional information, please see
    :ref:`resources-fonts-liberation`.

    .. list-table:: ``font_name`` values for :py:class:`arcade.Text`
       :header-rows: 1

       * - Proprietary Font(s)
         - Liberation Replacemetn

       * - ``"Courier"``
         - ``"Liberation Mono"``

       * - ``"Times New Roman"``, ``"Times"``
         - ``"Liberation Serif"``

       * - ``"Arial"``
         - ``"Liberation Sans"``

    """
    from arcade.text import load_font

    load_font(":system:fonts/ttf/Liberation/Liberation_Mono_BoldItalic.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Mono_Bold.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Mono_Italic.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Mono_Regular.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Sans_BoldItalic.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Sans_Bold.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Sans_Italic.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Sans_Regular.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Serif_BoldItalic.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Serif_Bold.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Serif_Italic.ttf")
    load_font(":system:fonts/ttf/Liberation/Liberation_Serif_Regular.ttf")



from dataclasses import dataclass
from typing import Any

from pyglet.math import Vec2


@dataclass
class UIEvent:
    """An event created by the GUI system.

    Can be passed as follows:

    .. code-block:: python

       widget.dispatch("on_event", event)

    An event always has a source. This is the UIManager for general input
    events, but will be the specific widget in case of events like click
    events.

    Args:
        source: The source of the event.
    """

    source: Any


@dataclass
class UIMouseEvent(UIEvent):
    """Base class for all UI mouse events.

    Args:
        x: The x coordinate of the mouse.
        y: The y coordinate of the mouse.
    """

    x: int
    y: int

    @property
    def pos(self) -> Vec2:
        """Return the position as tuple (x, y)"""
        return Vec2(self.x, self.y)


@dataclass
class UIMouseMovementEvent(UIMouseEvent):
    """Triggered when the mouse is moved.

    Args:
        dx: The change in x coordinate.
        dy: The change in y coordinate.
    """

    dx: int
    dy: int


@dataclass
class UIMousePressEvent(UIMouseEvent):
    """Triggered when a mouse button(left, right, middle) is pressed.

    Args:
        button: The button pressed.
        modifiers: The modifiers pressed.
    """

    button: int
    modifiers: int


@dataclass
class UIMouseDragEvent(UIMouseEvent):
    """Triggered when the mouse moves while one of its buttons being pressed.

    Args:
        dx: The change in x coordinate.
        dy: The change in y coordinate.
        buttons: The buttons pressed.
        modifiers: The modifiers pressed.
    """

    dx: int
    dy: int
    buttons: int
    modifiers: int


@dataclass
class UIMouseReleaseEvent(UIMouseEvent):
    """Triggered when a mouse button is released.

    Args:
        button: The button released.
        modifiers: The modifiers pressed
    """

    button: int
    modifiers: int


@dataclass
class UIMouseScrollEvent(UIMouseEvent):
    """Triggered by rotating the scroll wheel on the mouse.

    Args:
        scroll_x: The horizontal scroll amount.
        scroll_y: The vertical scroll
    """

    scroll_x: int
    scroll_y: int


@dataclass
class UIKeyEvent(UIEvent):
    """Base class for all keyboard-centric UI events.

    Args:
        symbol: The key pressed.
        modifiers: The modifiers pressed.
    """

    symbol: int
    modifiers: int


@dataclass
class UIKeyPressEvent(UIKeyEvent):
    """Triggered when a key is pressed."""

    pass


@dataclass
class UIKeyReleaseEvent(UIKeyEvent):
    """Triggered when a key is released."""

    pass


@dataclass
class UITextEvent(UIEvent):
    """Base class for text-related events.

    It holds no data of its own.
    """

    pass


@dataclass
class UITextInputEvent(UITextEvent):
    """Triggered whenever the user inputs any text.

    Usually, this will be after :py:class:`UIKeyPressEvent` and before a
    :py:class:`UIKeyReleaseEvent`. It may also occur when:

    * the user holds down a key to repeat letters or spaces
    * a platform-specific input method, such as pen input on a tablet PC

    To learn more, see pyglet's `relevant documentation <https://pyglet.readthedocs.io/en/development/modules/window.html#pyglet.window.Window.on_text>`_.

    Args:
        text: The text inputted. Often a single character.
    """

    text: str


@dataclass
class UITextMotionEvent(UITextEvent):
    """Triggered when text cursor moves.

    Args:
        motion: The motion of the cursor
    """

    motion: Any


@dataclass
class UITextMotionSelectEvent(UITextEvent):
    """Triggered when the text cursor moves selecting the text with it.

    Args:
        selection: The selection of the cursor
    """

    selection: Any


@dataclass
class UIOnClickEvent(UIMouseEvent):
    """Triggered when a widget is clicked.

    Args:
        button: The button clicked.
        modifiers: The modifiers pressed.
    """

    button: int
    modifiers: int


@dataclass
class UIOnUpdateEvent(UIEvent):
    """Arcade on_update callback passed as :class:`UIEvent`.

    Args:
        dt: Time since last update
    """

    dt: int


@dataclass
class UIOnChangeEvent(UIEvent):
    """Value of a widget changed.

    Args:
        old_value: The old value.
        new_value: The new value.
    """

    old_value: Any
    new_value: Any


@dataclass
class UIOnActionEvent(UIEvent):
    """Notification about an action.

    Args:
        action: Value describing the action, mostly a string
    """

    action: Any


@dataclass
class UIControllerEvent(UIEvent):
    """Base class for all UI controller events.

    Args:
        source: The controller that triggered the event.
    """


@dataclass
class UIControllerConnectEvent(UIControllerEvent):
    """Triggered when a controller is connected.

    Args:
        source: The controller that triggered the event.
    """


@dataclass
class UIControllerDisconnectEvent(UIControllerEvent):
    """Triggered when a controller is disconnected.

    Args:
        source: The controller that triggered the event.
    """


@dataclass
class UIControllerStickEvent(UIControllerEvent):
    """Triggered when a controller stick is moved.

    Args:
        name: The name of the stick.
        vector: The value of the stick.
    """

    name: str
    vector: Vec2


@dataclass
class UIControllerTriggerEvent(UIControllerEvent):
    """Triggered when a controller trigger is moved.

    Args:
        name: The name of the trigger.
        value: The value of the trigger.
    """

    name: str
    value: float


@dataclass
class UIControllerButtonEvent(UIControllerEvent):
    """Triggered when a controller button used.

    Args:
        button: The name of the button.
    """

    button: str


@dataclass
class UIControllerButtonPressEvent(UIControllerButtonEvent):
    """Triggered when a controller button is pressed.

    Args:
        button: The name of the button.
    """


@dataclass
class UIControllerButtonReleaseEvent(UIControllerButtonEvent):
    """Triggered when a controller button is released.

    Args:
        button: The name of the button.
    """


@dataclass
class UIControllerDpadEvent(UIControllerEvent):
    """Triggered when a controller dpad is moved.

    Args:
        vector: The value of the dpad.
    """

    vector: Vec2



"""
Platformer Template

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.template_platformer
"""
import arcade
from arcade.types import Color

# --- Constants
WINDOW_TITLE = "Platformer"
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720

# Constants used to scale our sprites from their original size
CHARACTER_SCALING = 1
TILE_SCALING = 0.5
COIN_SCALING = 0.5
SPRITE_PIXEL_SIZE = 128
GRID_PIXEL_SIZE = SPRITE_PIXEL_SIZE * TILE_SCALING

# Movement speed of player, in pixels per frame
PLAYER_MOVEMENT_SPEED = 10
GRAVITY = 1
PLAYER_JUMP_SPEED = 20

# Camera constants
FOLLOW_DECAY_CONST = 0.3  # get within 1% of the target position within 2 seconds


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        super().__init__()

        # A Camera that can be used for scrolling the screen
        self.camera_sprites = arcade.Camera2D()

        # A rectangle that is used to constrain the camera's position.
        # we update it when we load the tilemap
        self.camera_bounds = self.window.rect

        # A non-scrolling camera that can be used to draw GUI elements
        self.camera_gui = arcade.Camera2D()

        # The scene which helps draw multiple spritelists in order.
        self.scene = self.create_scene()

        # Set up the player, specifically placing it at these coordinates.
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_idle.png",
            scale=CHARACTER_SCALING,
        )

        # Our physics engine
        self.physics_engine = arcade.PhysicsEnginePlatformer(
            self.player_sprite, gravity_constant=GRAVITY, walls=self.scene["Platforms"]
        )

        # Keep track of the score
        self.score = 0

        # What key is pressed down?
        self.left_key_down = False
        self.right_key_down = False

        # Text object to display the score
        self.score_display = arcade.Text(
            "Score: 0",
            x=10,
            y=10,
            color=arcade.csscolor.WHITE,
            font_size=18,
        )

    def create_scene(self) -> arcade.Scene:
        """Load the tilemap and create the scene object."""
        # Our TileMap Object
        # Layer specific options are defined based on Layer names in a dictionary
        # Doing this will make the SpriteList for the platforms layer
        # use spatial hashing for collision detection.
        layer_options = {
            "Platforms": {
                "use_spatial_hash": True,
            },
        }
        tile_map = arcade.load_tilemap(
            ":resources:tiled_maps/map.json",
            scaling=TILE_SCALING,
            layer_options=layer_options,
        )

        # Set the window background color to the same as the map if it has one
        if tile_map.background_color:
            self.window.background_color = Color.from_iterable(tile_map.background_color)

        # Use the tilemap's size to correctly set the camera's bounds.
        # Because how how shallow the map is we don't offset the bounds height
        self.camera_bounds = arcade.LRBT(
            self.window.width/2.0,
            tile_map.width * GRID_PIXEL_SIZE - self.window.width/2.0,
            self.window.height/2.0,
            tile_map.height * GRID_PIXEL_SIZE
        )


        # Our Scene Object
        # Initialize Scene with our TileMap, this will automatically add all layers
        # from the map as SpriteLists in the scene in the proper order.
        return arcade.Scene.from_tilemap(tile_map)

    def reset(self):
        """Reset the game to the initial state."""
        self.score = 0
        # Load a fresh scene to get the coins back
        self.scene = self.create_scene()

        # Move the player to start position
        self.player_sprite.position = (128, 128)
        # Add the player to the scene
        self.scene.add_sprite("Player", self.player_sprite)

    def on_draw(self):
        """Render the screen."""

        # Clear the screen to the background color
        self.clear()

        # Draw the map with the sprite camera
        with self.camera_sprites.activate():
            # Draw our Scene
            # Note, if you a want pixelated look, add pixelated=True to the parameters
            self.scene.draw()

        # Draw the score with the gui camera
        with self.camera_gui.activate():
            # Draw our score on the screen. The camera keeps it in place.
            self.score_display.text = f"Score: {self.score}"
            self.score_display.draw()

    def update_player_speed(self):
        # Calculate speed based on the keys pressed
        self.player_sprite.change_x = 0

        if self.left_key_down and not self.right_key_down:
            self.player_sprite.change_x = -PLAYER_MOVEMENT_SPEED
        elif self.right_key_down and not self.left_key_down:
            self.player_sprite.change_x = PLAYER_MOVEMENT_SPEED

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed."""

        # Jump
        if key == arcade.key.UP or key == arcade.key.W:
            if self.physics_engine.can_jump():
                self.player_sprite.change_y = PLAYER_JUMP_SPEED

        # Left
        elif key == arcade.key.LEFT or key == arcade.key.A:
            self.left_key_down = True
            self.update_player_speed()

        # Right
        elif key == arcade.key.RIGHT or key == arcade.key.D:
            self.right_key_down = True
            self.update_player_speed()

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key."""
        if key == arcade.key.LEFT or key == arcade.key.A:
            self.left_key_down = False
            self.update_player_speed()
        elif key == arcade.key.RIGHT or key == arcade.key.D:
            self.right_key_down = False
            self.update_player_speed()

    def center_camera_to_player(self):
        # Move the camera to center on the player
        self.camera_sprites.position = arcade.math.smerp_2d(
            self.camera_sprites.position,
            self.player_sprite.position,
            self.window.delta_time,
            FOLLOW_DECAY_CONST,
        )

        # Constrain the camera's position to the camera bounds.
        self.camera_sprites.view_data.position = arcade.camera.grips.constrain_xy(
            self.camera_sprites.view_data, self.camera_bounds
        )

    def on_update(self, delta_time: float):
        """Movement and game logic"""

        # Move the player with the physics engine
        self.physics_engine.update()

        # See if we hit any coins
        coin_hit_list = arcade.check_for_collision_with_list(
            self.player_sprite, self.scene["Coins"]
        )

        # Loop through each coin we hit (if any) and remove it
        for coin in coin_hit_list:
            # Remove the coin
            coin.remove_from_sprite_lists()
            # Add one to the score
            self.score += 1

        # Position the camera
        self.center_camera_to_player()

    def on_resize(self, width: int, height: int):
        """ Resize window """
        super().on_resize(width, height)
        # Update the cameras to match the new window size
        self.camera_sprites.match_window()
        # The position argument keeps `0, 0` in the bottom left corner.
        self.camera_gui.match_window(position=True)


def main():
    """Main function"""
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    game = GameView()
    game.reset()

    window.show_view(game)
    arcade.run()


if __name__ == "__main__":
    main()



"""
See: https://www.shadertoy.com/view/lsBfRc
"""

from arcade.experimental import Shadertoy


class BloomFilter:
    """
    CRT Filter

    Args:
        width: Width of the filter
        height: Height of the filter
    """

    def __init__(self, width: int, height: int, intensity: float):
        self.shadertoy = Shadertoy.create_from_file(
            (width, height), ":system:shaders/bloom/bloom_filter_image.glsl"
        )
        self.fbo = self.shadertoy.ctx.framebuffer(
            color_attachments=[self.shadertoy.ctx.texture((width, height), components=4)]
        )

        self.shadertoy.buffer_a = self.shadertoy.create_buffer_from_file(
            ":system:shaders/bloom/bloom_filter_a.glsl"
        )
        self.shadertoy.buffer_a.program["intensity"] = intensity
        self.shadertoy.buffer_a.channel_0 = self.fbo.color_attachments[0]

        self.shadertoy.buffer_b = self.shadertoy.create_buffer_from_file(
            ":system:shaders/bloom/bloom_filter_b.glsl"
        )
        self.shadertoy.buffer_b.channel_0 = self.shadertoy.buffer_a.texture

        self.shadertoy.channel_0 = self.shadertoy.buffer_a.texture
        self.shadertoy.channel_1 = self.shadertoy.buffer_b.texture

    def use(self):
        self.fbo.use()

    def clear(self):
        self.fbo.clear()

    def draw(self, time: float = 0, target=None):
        self.shadertoy.render()



"""Example of creating a custom progress bar.

This example shows how to create a custom progress bar.

A progress bar consists of a black background box and a color bar that fills the box
from left to right. Unfilled space is gray.
A value between 0 and 1 determines the fill level of the bar.

The first progress bar is created using a `UIWidget` and the second progress bar is
created using a `UIAnchorLayout`.

For the first approach, you only need to know about the general GUI concepts, specifically
how widgets are rendered.

For the second approach, you need to know how to use layouts and size_hints to arrange widgets.

Both approaches use a Property to trigger a render when the value changes.
Properties are a way to bind a value to a widget and trigger a function when the value changes.
Read more about properties in the `arcade.gui` documentation.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.own_widgets
"""

import arcade
from arcade.gui import Property, UIAnchorLayout, UIBoxLayout, UISpace, UIView, UIWidget, bind
from arcade.types import Color


class ProgressBar1(UIWidget):
    """A custom progress bar widget.

    A UIWidget is a basic building block for GUI elements. It is a rectangle with a
    background color and can have children.

    To create a custom progress bar, we create a UIWidget with a black background,
    set a border and add a `do_render` method to draw the actual progress bar.

    """

    value = Property(0.0)
    """The fill level of the progress bar. A value between 0 and 1."""

    def __init__(
        self,
        *,
        value: float = 1.0,
        width=100,
        height=20,
        color: Color = arcade.color.GREEN,
    ) -> None:
        super().__init__(
            width=width,
            height=height,
            size_hint=None,  # disable size hint, so it just uses the size given
        )
        self.with_background(color=arcade.uicolor.GRAY_CONCRETE)
        self.with_border(color=arcade.uicolor.BLACK)

        self.value = value
        self.color = color

        # trigger a render when the value changes
        bind(self, "value", self.trigger_render)

    def do_render(self, surface: arcade.gui.Surface) -> None:
        """Draw the actual progress bar."""
        # this will set the viewport to the size of the widget
        # so that 0,0 is the bottom left corner of the widget content
        self.prepare_render(surface)

        # Draw the actual bar
        arcade.draw_lbwh_rectangle_filled(
            0,
            0,
            self.content_width * self.value,
            self.content_height,
            self.color,
        )


class Progressbar2(UIAnchorLayout):
    """A custom progress bar widget.

    A UIAnchorLayout is a layout that arranges its children in a specific way.
    The actual bar is a UISpace that fills the parent widget from left to right.
    """

    value = Property(0.0)

    def __init__(
        self,
        value: float = 1.0,
        width=100,
        height=20,
        color: Color = arcade.color.GREEN,
    ) -> None:
        super().__init__(
            width=width,
            height=height,
            size_hint=None,  # disable size hint, so it just uses the size given
        )
        self.with_background(color=arcade.uicolor.GRAY_CONCRETE)
        self.with_border(color=arcade.uicolor.BLACK)

        self._bar = UISpace(
            color=color,
            size_hint=(value, 1),
        )
        self.add(
            self._bar,
            anchor_x="left",
            anchor_y="top",
        )
        self.value = value

        # update the bar when the value changes
        bind(self, "value", self._update_bar)

    def _update_bar(self):
        self._bar.size_hint = (self.value, 1)
        self._bar.visible = self.value > 0


class MyView(UIView):
    def __init__(self):
        super().__init__()
        self.background_color = arcade.uicolor.BLUE_BELIZE_HOLE

        root = self.ui.add(UIAnchorLayout())
        bars = root.add(UIBoxLayout(space_between=10))

        # UIWidget based progress bar
        self.progressbar1 = ProgressBar1(
            value=0.8,
            color=arcade.color.RED,
        )
        bars.add(self.progressbar1)

        # UIAnchorLayout based progress bar
        self.progressbar2 = Progressbar2(
            value=0.8,
            color=arcade.color.BLUE,
        )
        bars.add(self.progressbar2)

    def on_key_press(self, symbol: int, modifiers: int) -> bool | None:
        if symbol == arcade.key.NUM_ADD:
            self.progressbar1.value = (self.progressbar1.value + 0.1) % 1
            self.progressbar2.value = (self.progressbar2.value + 0.1) % 1
        elif symbol == arcade.key.NUM_SUBTRACT:
            self.progressbar1.value = (self.progressbar1.value - 0.1) % 1
            self.progressbar2.value = (self.progressbar2.value - 0.1) % 1

        return None


def main():
    window = arcade.Window(title="GUI Example: Progressbar")
    window.show_view(MyView())
    arcade.run()


if __name__ == "__main__":
    main()



"""
This submodule has functions that control opening, closing, rendering,
and otherwise managing windows.
It also has commands for scheduling pauses and scheduling interval functions.
"""

from __future__ import annotations

import gc
import os
import time
from collections.abc import Callable
from typing import TYPE_CHECKING

import pyglet

from arcade.types import RGBA255, Color

if TYPE_CHECKING:
    from arcade import Window
    from arcade.application import View

_window: Window | None = None

__all__ = [
    "get_display_size",
    "get_window",
    "set_window",
    "close_window",
    "run",
    "exit",
    "start_render",
    "finish_render",
    "set_background_color",
    "schedule",
    "unschedule",
    "schedule_once",
]


def get_display_size(screen_id: int = 0) -> tuple[int, int]:
    """
    Return the width and height of a monitor.

    The size of the primary monitor is returned by default.

    Args:
        screen_id: The screen number
    Returns:
        Tuple containing the width and height of the screen
    """
    display = pyglet.display.Display()  # type: ignore  # pending: pyglet fixes import tricks
    screen = display.get_screens()[screen_id]
    return screen.width, screen.height


def get_window() -> Window:
    """
    Return a handle to the current window.

    :return: Handle to the current window.
    """
    if _window is None:
        raise RuntimeError("No window is active. It has not been created yet, or it was closed.")

    return _window


def set_window(window: Window | None) -> None:
    """
    Set a handle to the current window.

    Args:
        window: Handle to the current window.
    """
    global _window
    _window = window


def close_window() -> None:
    """
    Closes the current window, and then runs garbage collection. The garbage collection
    is necessary to prevent crashing when opening/closing windows rapidly (usually during
    unit tests).
    """
    global _window

    if _window is None:
        return

    _window.close()
    _window = None

    # Have to do a garbage collection or Python will crash
    # if we do a lot of window open and closes. Like for
    # unit tests.
    gc.collect()


def run(view: View | None = None) -> None:
    """
    Run the main loop. Optionally start with a specified view.

    After the window has been set up, and the event hooks are in place,
    this is usually one of the last commands on the main program.
    This is a blocking function starting pyglet's event loop meaning
    it will start to dispatch events such as ``on_draw`` and ``on_update``.

    This function also handles special cases for running in headless mode
    and running unit tests.

    Args:
        view: The view to display when starting the run. Defaults to None.
    """
    window = get_window()

    if view is not None:
        window.show_view(view)

    # Some unite tests needs to run a single draw/update cycle
    if os.environ.get("ARCADE_TEST"):
        window.on_update(1.0 / 60.0)
        window.on_draw()
    elif window.headless:
        # We are entering headless more an will emulate an event loop

        # Ensure the initial delta time is not 0 to be
        # more in line with how a normal window works.
        delta_time = window._draw_rate
        last_time = time.perf_counter()

        # As long as we have a context --
        while window.context:
            # Select active view or window
            active = window.current_view or window

            active.on_update(delta_time)
            if window.context:
                active.on_draw()

            # Window could be closed in on_draw
            if window.context:
                window.flip()

            now = time.perf_counter()
            delta_time, last_time = now - last_time, now
    else:
        # Start the standard event loop (blocking)
        # Note that we pass None as the interval here because we register
        # a single interval event to dispatch frames because separate interval
        # events cause serious issues with framerate smoothness.
        pyglet.app.run(None)


def exit() -> None:
    """
    Exits the application.
    """
    pyglet.app.exit()


def start_render(pixelated=False, blend=True) -> None:
    """
    Start recording drawing functions into an offscreen buffer.
    Call :py:func:`arcade.finish_render` to stop recording. The
    start_render/finish_render calls can only be called once.

    When running Arcade this buffer will be presented to the screen.

    A few configuration options are available in this function.

    Args:
        pixelated:
            If True, the buffer will be be pixelated when resized.
            Otherwise, it will be smooth.
        blend:
            If alpha blending
    """
    from arcade.start_finish_data import StartFinishRenderData

    window = get_window()
    if window._start_finish_render_data is not None:
        raise RuntimeError(
            "start_render() can only be called once during the application's lifetime "
            "and should only be used when calling draw functions at module level in "
            "a simple script to produce a static image. If you are seeing this error "
            "you likely intended to call clear() instead."
        )

    window._start_finish_render_data = StartFinishRenderData(pixelated=pixelated, blend=blend)
    window._start_finish_render_data.begin()


def finish_render() -> None:
    """
    Stop recording drawing functions into an offscreen buffer.
    :py:func:`arcade.start_render` should be called before this function.

    :py:func:`arcade.run` can be called after this function to present the buffer.
    """
    window = get_window()
    if window._start_finish_render_data is None:
        raise RuntimeError("finish_render() was called without a matching start_render() call.")

    if window._start_finish_render_data.completed:
        raise RuntimeError("finish_render() was called more than once.")

    window._start_finish_render_data.end()


def set_background_color(color: RGBA255) -> None:
    """
    Set the color :py:meth:`arcade.Window.clear()` will use
    when clearing the window. This only needs to be called
    when the background color changes.

    .. Note::

        A shorter and faster way to set background color
        is using :py:attr:`arcade.Window.background_color`.

    Examples::

        # Use Arcade's built in color values
        arcade.set_background_color(arcade.color.AMAZON)

        # Specify RGB value directly (red)
        arcade.set_background_color((255, 0, 0))

    Args:
        color: List of 3 or 4 values in RGB/RGBA format.
    """
    get_window().background_color = Color.from_iterable(color)


def schedule(function_pointer: Callable, interval: float):
    """
    Schedule a function to be automatically called every ``interval``
    seconds. The function/callable needs to take a delta time argument
    similar to ``on_update``. This is a float representing the number
    of seconds since it was scheduled or called.

    A function can be scheduled multiple times, but this is not recommended.

    .. Warning:: Scheduled functions should **always** be unscheduled
                 using :py:func:`arcade.unschedule`. Having lingering
                 scheduled functions will lead to crashes.

    Example::

        def some_action(delta_time):
            print(delta_time)

        # Call the function every second
        arcade.schedule(some_action, 1)
        # Unschedule

    Args:
        function_pointer:
            Pointer to the function to be called.
        interval:
            Interval to call the function (float or integer)
    """
    pyglet.clock.schedule_interval(function_pointer, interval)


def unschedule(function_pointer: Callable):
    """
    Unschedule a function being automatically called.

    Example::

        def some_action(delta_time):
            print(delta_time)

        arcade.schedule(some_action, 1)
        arcade.unschedule(some_action)

    Args:
        function_pointer: Pointer to the function to be unscheduled.
    """
    pyglet.clock.unschedule(function_pointer)


def schedule_once(function_pointer: Callable, delay: float):
    """
    Schedule a function to be automatically called once after ``delay``
    seconds. The function/callable needs to take a delta time argument
    similar to ``on_update``. This is a float representing the number
    of seconds since it was scheduled or called.

    Example::

        def some_action(delta_time):
            print(delta_time)

        # Call the function once after 1 second
        arcade.schedule_once(some_action, 1)

    Args:
        function_pointer:
            Pointer to the function to be called.
        delay:
            Delay in seconds
    """
    pyglet.clock.schedule_once(function_pointer, delay)



"""
Example showing how to use the easing functions for position.
Example showing how to use easing for angles.

See:
https://easings.net/
...for a great guide on the theory behind how easings can work.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.easing_example_2
"""

import arcade
from arcade import easing

SPRITE_SCALING = 1.0

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Easing Example"


class Player(arcade.Sprite):
    """Player class"""

    def __init__(self, image, scale):
        """Set up the player"""

        # Call the parent init
        super().__init__(image, scale=scale)

        self.easing_angle_data = None
        self.easing_x_data = None
        self.easing_y_data = None

    def update(self, delta_time: float = 1 / 60):
        if self.easing_angle_data is not None:
            done, self.angle = easing.ease_angle_update(self.easing_angle_data, delta_time)
            if done:
                self.easing_angle_data = None

        if self.easing_x_data is not None:
            done, self.center_x = easing.ease_update(self.easing_x_data, delta_time)
            if done:
                self.easing_x_data = None

        if self.easing_y_data is not None:
            done, self.center_y = easing.ease_update(self.easing_y_data, delta_time)
            if done:
                self.easing_y_data = None


class GameView(arcade.View):
    """Main application class."""

    def __init__(self):
        """Initializer"""

        # Call the parent class initializer
        super().__init__()

        # Set up the player info
        self.player_list = arcade.SpriteList()

        # Load the player texture. The ship points up by default. We need it to point right.
        # That's why we rotate it 90 degrees clockwise.
        texture = arcade.load_texture(":resources:images/space_shooter/playerShip1_orange.png")
        texture = texture.rotate_90()

        # Set up the player
        self.player_sprite = Player(texture, SPRITE_SCALING)
        self.player_sprite.angle = 0
        self.player_sprite.center_x = WINDOW_WIDTH / 2
        self.player_sprite.center_y = WINDOW_HEIGHT / 2
        self.player_list.append(self.player_sprite)

        # Set the background color
        self.background_color = arcade.color.BLACK
        self.text = "Move the mouse and press 1-9 to apply an easing function."

    def on_draw(self):
        """Render the screen."""

        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.player_list.draw()

        arcade.draw_text(self.text, 15, 15, arcade.color.WHITE, 24)

    def on_update(self, delta_time):
        """Movement and game logic"""

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.player_list.update(delta_time)

    def on_key_press(self, key, modifiers):
        x = self.window.mouse["x"]
        y = self.window.mouse["y"]

        if key == arcade.key.KEY_1:
            angle = arcade.math.get_angle_degrees(
                x1=self.player_sprite.position[0], y1=self.player_sprite.position[1], x2=x, y2=y
            )
            self.player_sprite.angle = angle
            self.text = "Instant angle change"
        if key in [arcade.key.KEY_2, arcade.key.KEY_3, arcade.key.KEY_4, arcade.key.KEY_5]:
            p1 = self.player_sprite.position
            p2 = (x, y)
            end_angle = arcade.math.get_angle_degrees(p1[0], p1[1], p2[0], p2[1])
            start_angle = self.player_sprite.angle
            if key == arcade.key.KEY_2:
                ease_function = easing.linear
                self.text = "Linear easing - angle"
            elif key == arcade.key.KEY_3:
                ease_function = easing.ease_in
                self.text = "Ease in - angle"
            elif key == arcade.key.KEY_4:
                ease_function = easing.ease_out
                self.text = "Ease out - angle"
            elif key == arcade.key.KEY_5:
                ease_function = easing.smoothstep
                self.text = "Smoothstep - angle"
            else:
                raise ValueError("?")

            self.player_sprite.easing_angle_data = easing.ease_angle(
                start_angle, end_angle, rate=180, ease_function=ease_function
            )

        if key in [arcade.key.KEY_6, arcade.key.KEY_7, arcade.key.KEY_8, arcade.key.KEY_9]:
            p1 = self.player_sprite.position
            p2 = (x, y)
            if key == arcade.key.KEY_6:
                ease_function = easing.linear
                self.text = "Linear easing - position"
            elif key == arcade.key.KEY_7:
                ease_function = easing.ease_in
                self.text = "Ease in - position"
            elif key == arcade.key.KEY_8:
                ease_function = easing.ease_out
                self.text = "Ease out - position"
            elif key == arcade.key.KEY_9:
                ease_function = easing.smoothstep
                self.text = "Smoothstep - position"
            else:
                raise ValueError("?")

            ex, ey = easing.ease_position(p1, p2, rate=180, ease_function=ease_function)
            self.player_sprite.easing_x_data = ex
            self.player_sprite.easing_y_data = ey

    def on_mouse_press(self, x: float, y: float, button: int, modifiers: int):
        angle = arcade.math.get_angle_degrees(
            x1=self.player_sprite.position[0], y1=self.player_sprite.position[1], x2=x, y2=y
        )
        self.player_sprite.angle = angle


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from pathlib import Path

import arcade
from arcade.experimental.shadertoy import Shadertoy


class MyGame(arcade.Window):
    def __init__(self, width, height, title):
        super().__init__(width, height, title, resizable=True)
        self.shadertoy = Shadertoy.create_from_file(
            self.get_framebuffer_size(), Path(__file__).parent / "star_nest.glsl"
        )

    def on_draw(self):
        self.clear()
        self.shadertoy.render()

    def on_update(self, dt):
        # Update the internal time in shadertoy
        self.shadertoy.time += dt

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        self.shadertoy.mouse_position = x, y

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.shadertoy.resize((width, height))


if __name__ == "__main__":
    MyGame(1280, 720, "Shadertoy Demo")
    arcade.run()



import math
import time

import arcade

# Do the math to figure out our screen dimensions
SCREEN_WIDTH = 1280
SCREEN_HEIGHT = 720
SCREEN_TITLE = "Basic Renderer"


class MyGame(arcade.Window):
    def __init__(self, width, height, title):
        super().__init__(width, height, title)
        # vsync must be off when measuring rendering calls
        self.set_vsync(False)

        start = time.time()
        self.sprites = arcade.SpriteList()
        num_sprites = 100_000
        sprite_scale = 0.02  # All sprites covering the screen
        # sprite_scale = 1.0  # default
        self.texture = arcade.load_texture(":assets:images/tiles/boxCrate.png")

        r = int(math.sqrt(num_sprites))
        for y in range(r):
            for x in range(r):
                self.sprites.append(
                    arcade.Sprite(
                        self.texture,
                        scale=sprite_scale,
                        center_x=128 * sprite_scale * x,
                        center_y=128 * sprite_scale * y,
                    )
                )
        self.sprites.draw()  # Force the list to build

        self.sprites.program = self.ctx.sprite_list_program_no_cull
        print(f"Initialization time: {time.time() - start}")

        self.query = self.ctx.query()
        self.frames = 0
        self.frame_step = 60
        self.time_elapsed_total = 0

    def on_draw(self):
        self.clear()

        with self.query:
            self.sprites.draw()

        self.time_elapsed_total += self.query.time_elapsed
        self.frames += 1

        if self.frames % self.frame_step == 0:
            print(f"--- Stats over {self.frame_step} frames")
            print(f"Time elapsed       : {self.time_elapsed_total / 1_000_000_000} seconds")
            print(f"Samples passed     : {self.query.samples_passed}")
            print(f"Primitives created : {self.query.primitives_generated}")
            self.time_elapsed_total = 0


if __name__ == "__main__":
    MyGame(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_TITLE)
    arcade.run()
